@@ -9,13 +9,14 @@
  *  Contributors:
  *    Denys Duchier (duchier@ps.uni-sb.de)
  *    Per Brand (perbrand@sics.se)
+ *    Kevin Glynn (glynn@info.ucl.ac.be)
  * 
  *  Copyright:
  *    Organization or Person (Year(s))
  * 
  *  Last change:
- *    $Date: 2002/11/24 23:12:04 $ by $Author: duchier $
- *    $Revision: 1.66.2.1.2.2 $
+ *    $Date: 2004/04/06 12:27:38 $ by $Author: popow $
+ *    $Revision: 1.76 $
  * 
  *  This file is part of Mozart, an implementation 
  *  of Oz 3:
@@ -42,7 +43,8 @@
 #include "var_bool.hh"
 #include "var_of.hh"
 #include "var_ct.hh"
-#include "var_future.hh"
+#include "var_failed.hh"
+#include "var_readonly.hh"
 #include "var_simple.hh"
 #include "var_opt.hh"
 #include "var_ext.hh"
@@ -55,10 +57,31 @@
 #include "weakdict.hh"
 #include "mozart_cpi.hh"
 
+#if defined(DEBUG_LIVECALC)
+// Whenever liveness zeroes the contents of a register we actually
+// place a fresh int (maintained by deadMarker) there. This gives some 
+// debugging info when tracking down buggy liveness code ...
+static int deadMarker = 300000;
+   #define VOID_GREG(to,i) to->initG(i,makeTaggedSmallInt(deadMarker++))
+#else
+   #define VOID_GREG(to,i) to->initG(i,makeTaggedNULL())
+#endif
+#if defined(DEBUG_PRINTLIVECALC)
+static int deadMarker = makeTaggedNULL();
+#endif
+
 #ifdef OUTLINE
 #define inline
 #endif
 
+
+/* We allocate a static usage block for G usage vectors.  This
+** saves a lot of malloc/free and a lot of time for some benchmarks
+*/
+#define StaticGUsageVectorSize 100
+
+static int gUsageVector[StaticGUsageVectorSize];
+
 /*
  * Do some bad style renaming
  *
@@ -102,6 +125,8 @@
 #define _cacSuspList             gCollectSuspList
 #define _cacLocalSuspList        gCollectLocalSuspList
 
+#define _cacDictEntry		gCollectDictEntry
+
 #endif
 
 
@@ -143,6 +168,8 @@
 #define _cacSuspList             sCloneSuspList
 #define _cacLocalSuspList        sCloneLocalSuspList
 
+#define _cacDictEntry		sCloneDictEntry
+
 #endif
 
 
@@ -287,7 +314,7 @@ void exitCheckSpace() {
     _t[0]=_f[0];                                                             \
     break;                                                                   \
   default:                                                                   \
-    memcpy(_t,_f,_n);                                                        \
+    memcpy(_t, _f, _n);                                                      \
   }                                                                          \
   t = (Type *) _t;                                                           \
 }
@@ -336,6 +363,7 @@ void exitCheckSpace() {
  *
  */
 
+inline
 RefsArray * RefsArray::_cac(void) {
   if (!this)
     return this;
@@ -374,6 +402,123 @@ RefsArray * RefsArray::_cac(void) {
 #endif
 
 
+#ifdef G_COLLECT
+/*
+ * Abstraction
+ * Argument vector describes liveness of G registers. If 
+ * it is NULL then all registers are live.
+ */
+Abstraction * Abstraction::gCollect(int gUsageLen, int * gUsage) {
+
+  Assert(this);
+  if (cacIsMarked())
+    return (Abstraction *) cacGetFwd();
+
+  Abstraction *to;
+  TaggedRef *gReg;
+  int gFinished = 1;
+
+  if (cacIsCopied()) {
+    // Repeat customer, update and gc any newly live fields in the copy.
+    to = cacGetCopy();
+    gReg = to->getGRef();
+    if (!gUsage) {
+      // All previously dead registers are live
+      for (int i=to->pred->getGSize(); i--;) {
+ 	// Only gc if this slot was previously void
+	// to avoid nasty cycles.
+#if defined(DEBUG_LIVECALC)
+ 	if ( oz_isSmallInt(to->getG(i)) ) 
+#else
+	if ( to->getG(i)==makeTaggedNULL() ) 
+#endif
+	{
+	  to->initG(i,getG(i));
+	  oz_cacTerm(gReg[i], gReg[i]);
+	}
+      }
+    }
+    else {
+      for (int i=to->pred->getGSize(); i--;) {
+#if defined(DEBUG_LIVECALC)
+        if ( oz_isSmallInt(to->getG(i)) ) 
+#else
+	if ( to->getG(i)==makeTaggedNULL() ) 
+#endif
+	{ // Was dead
+	  if (gUsage[i]) {
+	    // Live. Only gc if this slot was previously void
+	    // to avoid nasty cycles.
+	    to->initG(i,getG(i));
+	    oz_cacTerm(gReg[i], gReg[i]);
+	  }
+	  else
+	    gFinished = 0;
+	}
+      }
+    }
+    if (gFinished) {
+      // Can skip this abstraction in future
+      STOREFWDFIELD(this, to);
+    }
+    return to;
+  }
+  // First time seen
+  to = (Abstraction *) oz_hrealloc(this, getAllocSize());
+  cacCopy(to);
+
+  gReg = to->getGRef();
+  int gSize= to->pred->getGSize();
+
+  // Void dead registers
+  if (gUsage) {
+    // *IMPORTANT*: must zero all dead regs before we do *any* gc
+    for (int i=gSize; i--; ) {
+      if (!(gUsage[i])) {
+#if defined(DEBUG_LIVECALC) || defined(DEBUG_PRINTLIVECALC)
+	printf("Abstraction <%s> G[%d] => %d\n", to->getPrintName(), i, deadMarker);
+#endif
+	VOID_GREG(to,i);
+        // Must still consider other paths to this abstraction
+	gFinished = 0;
+      }
+    }
+  }
+
+  if (gFinished) {
+    // Can skip this abstraction in future
+    STOREFWDFIELD(this, to);
+  }
+
+  if (to->hasGName())
+    gCollectGName(to->getGName1());
+  else
+    to->setBoard(to->getSubBoardInternal()->gCollectBoard());
+
+  // We must be careful to only collect the non-voided entries.
+  // Originally I fell through to the OZ_cacBlock call but that 
+  // caused a (hard to track down) bug whereby a register which is currently
+  // void could already have a collected entry in it by the time we get to it
+  // If gc'ing one of the registers causes this closure to be gc'd again the
+  // code in the cacIsCopied() branch ensures it doesn't touch 
+  // any of the registers we are gc'ing here.
+  if (!gFinished) {
+    for (int i=gSize; i--; ) {
+      if (gUsage[i]) {
+	oz_cacTerm(gReg[i], gReg[i]);
+      }
+    }
+  }
+  else // All registers still live.
+    OZ_cacBlock(gReg,gReg,gSize);
+
+  //cause the rest of the closure to be gc'd.
+  cacStack.push(to, PTR_CONSTTERM);
+  return to;
+}
+
+#endif /* G_COLLECT */
+
 /*
  * Boards:
  *
@@ -429,7 +574,7 @@ Name *Name::_cacName(void) {
 
   if (cacIsMarked())
     return cacGetFwd();
-  
+
 #ifdef G_COLLECT
   GName * gn = NULL;
 
@@ -438,12 +583,13 @@ Name *Name::_cacName(void) {
   }
 #endif
 
+  if (
 #ifdef G_COLLECT
-  if (isOnHeap()) {
+      isOnHeap()
 #else
-  if (!getBoardInternal()->hasMark()) {
+      !getBoardInternal()->hasMark()
 #endif
-
+      ) {
     Name * aux = (Name *) memcpy(oz_heapDoubleMalloc(sizeof(Name)), 
 				 this, sizeof(Name));
 
@@ -477,7 +623,6 @@ Literal * Literal::_cac(void) {
   return ((Name*) this)->_cacName();
 }
 
-
 /*
  * Dynamic tables
  *
@@ -498,6 +643,178 @@ DynamicTable * DynamicTable::_cac(void) {
   return to;
 }
 
+//
+// OzDictionary"s tables;
+//
+
+// replicated from dictionary.cc;
+static const double GDT_MAXENTRIES	= 0.9;
+static const double GDT_IDEALENTRIES	= 0.7;
+static const int GDT_MINFULL		= 4;   // 
+
+// This one does not change the 'entries' counter (nor calls
+// 'resize()', of course);
+void DictHashTable::_cacDictEntry(DictNode *n)
+{
+  DictNode *np = &table[hash(featureHash(n->getKey()))];
+
+  //
+  if (np->isEmpty()) {
+    (void) new (np) DictNode(*n);
+    OZ_cacBlock((TaggedRef *) n, (TaggedRef *) np, 2);
+    return;
+
+    //
+  } else {
+    if (!np->isPointer()) {
+      Assert(!featureEq(np->getKey(), n->getKey())); // unique keys;
+      // a fresh collision - allocate a new block of DictNode"s;
+      DictNode *newA =
+	(DictNode *) CAC_MALLOC(2 * sizeof(DictNode));
+
+      // .. this one must be already GC"ed;
+      (void) new (newA) DictNode(*np);
+      np->setSPtr(newA++);
+      (void) new (newA) DictNode(*n);
+      OZ_cacBlock((TaggedRef *) n, (TaggedRef *) newA, 2);
+      //
+      newA++;
+      np->setEPtr(newA);
+      return;
+
+      //
+    } else {
+      DictNode *sptr = np->getDictNodeSPtr();
+      DictNode *eptr = np->getDictNodeEPtr();
+      int bytes = ((char *) eptr) - ((char *) sptr) + sizeof(DictNode);
+      DictNode *newA = (DictNode *) CAC_MALLOC(bytes);
+
+      //
+      np->setSPtr(newA);
+      do {
+	(void) new (newA++) DictNode(*sptr++);
+      } while (sptr < eptr);
+      (void) new (newA) DictNode(*n);
+      OZ_cacBlock((TaggedRef *) n, (TaggedRef *) newA, 2);
+      //
+      newA++;
+      np->setEPtr(newA);
+      return;
+    }
+  }
+  Assert(0);
+}
+
+inline
+DictHashTable* DictHashTable::_cac(void)
+{
+  int tableSize = dictHTSizes[sizeIndex];
+  DictNode *an;
+
+  //
+  Assert(entries >= 0);
+  if (entries >= (tableSize / GDT_MINFULL)) {
+    // no compactification - reconstruct it isomorphically;
+    an = (DictNode *) CAC_MALLOC(tableSize * sizeof(DictNode));
+    DebugCode(int eS = 0;);
+    DebugCode(int oS = 0;);
+    DebugCode(int pS = 0;);
+
+    //
+    for (int i = tableSize; i--; ) {
+      DictNode *n = &table[i];
+      if (!n->isEmpty()) {
+	if (!n->isPointer()) {
+	  DebugCode(oS++);
+	  DictNode *nn = &an[i];
+	  (void) new (nn) DictNode(*n);
+	  OZ_cacBlock((TaggedRef *) n, (TaggedRef *) nn, 2);
+	} else {
+	  DebugCode(pS++);
+	  DictNode *sptr = n->getDictNodeSPtr();
+	  DictNode *eptr = n->getDictNodeEPtr();
+	  int bytes = ((char *) eptr) - ((char *) sptr);
+	  DictNode *newA = 
+	    (DictNode *) memcpy(CAC_MALLOC(bytes), sptr, bytes);
+	  int size = bytes / sizeof(DictNode);
+	  OZ_cacBlock((TaggedRef *) sptr, (TaggedRef *) newA, 2 * size);
+	  DictNode *nn = &an[i];
+	  nn->setSPtr(newA);
+	  nn->setEPtr((DictNode *) (((char *) newA) + bytes));
+	}
+      } else {
+	(void) new (&an[i]) DictNode; // empty;
+	DebugCode(eS++);
+      }
+    }
+
+    //
+#if defined(DEBUG_CHECK)
+//      if (tableSize >= 10000) {
+//        fprintf(stdout,
+//  	      "=(pid=%d)tabSize=%d, entries=%d, empty=%d, ord=%d, pts=%d\n",
+//  	      osgetpid(), tableSize, entries, eS, oS, pS);
+//      }
+#endif
+    DictHashTable *dht = new DictHashTable(*this);
+    dht->table = an;
+    return (dht);
+
+    //
+  } else {
+    // construct anew, GC"ing keys/values along;
+    //
+    int oldSize, newSize;
+    DictNode* old;
+
+    //
+    oldSize = dictHTSizes[sizeIndex];
+    old = table;
+
+    // can be zero too:
+    int tableSize = (int) ((double) entries * GDT_IDEALENTRIES);
+    Assert(tableSize < oldSize);
+    sizeIndex--;
+    while (sizeIndex >= 0 && dictHTSizes[sizeIndex] >= tableSize)
+      sizeIndex--;
+    Assert(sizeIndex < 0 || dictHTSizes[sizeIndex] < tableSize);
+    sizeIndex++;
+    Assert(sizeIndex >= 0 && dictHTSizes[sizeIndex] >= tableSize);
+    // Must not oscillate:
+    Assert(dictHTSizes[sizeIndex] < oldSize);
+    // Next GC should not attempt compactification:
+    Assert(entries >= (dictHTSizes[sizeIndex] / GDT_MINFULL));
+
+    // construct the table anew (keep the 'entries' counter);
+    tableSize = dictHTSizes[sizeIndex];
+    maxEntries = (int) (GDT_MAXENTRIES * tableSize);
+    table = (DictNode *) oz_heapMalloc(tableSize * sizeof(DictNode));
+    for (int i = tableSize; i--; )
+      (void) new (&table[i]) DictNode;
+
+    //
+    for (int i = oldSize; i--; old++) {
+      if (!old->isEmpty()) {
+	if (!old->isPointer()) {
+	  _cacDictEntry(old);
+	} else {
+	  DictNode *sptr = old->getDictNodeSPtr();
+	  DictNode *eptr = old->getDictNodeEPtr();
+	  do {
+	    _cacDictEntry(sptr++);
+	  } while (sptr < eptr);
+	}
+      }
+    }
+
+    //      
+    DictHashTable *dht = new DictHashTable(*this);
+    dht->table = table;
+    return (dht);
+  }
+  Assert(0);
+}
+
 /*
  * Variables
  *
@@ -563,6 +880,7 @@ OzVariable * OzVariable::_cacVarInline(void) {
   switch (getType()) {
   case OZ_VAR_OPT:
     Assert(0);
+  case OZ_VAR_SIMPLE_QUIET:
   case OZ_VAR_SIMPLE:
     cacReallocStatic(OzVariable,this,to,sizeof(SimpleVar));
     break;
@@ -572,8 +890,12 @@ OzVariable * OzVariable::_cacVarInline(void) {
     GCDBG_INTOSPACE(to);
     cacStack.push(to, PTR_VAR);
     break;
-  case OZ_VAR_FUTURE:
-    cacReallocStatic(OzVariable,this,to,sizeof(Future));
+  case OZ_VAR_READONLY_QUIET:
+  case OZ_VAR_READONLY:
+    cacReallocStatic(OzVariable,this,to,sizeof(ReadOnly));
+    break;
+  case OZ_VAR_FAILED:
+    cacReallocStatic(OzVariable,this,to,sizeof(Failed));
     cacStack.push(to, PTR_VAR);
     break;
   case OZ_VAR_BOOL:
@@ -613,16 +935,16 @@ void OzOFVariable::_cacRecurse(void) {
 }
 
 inline
-void Future::_cacRecurse(void) {
-  oz_cacTerm(function,function);
+void Failed::_cacRecurse(void) {
+  oz_cacTerm(exception,exception);
 }
 
 inline
 void OzVariable::_cacVarRecurse(void) {
   
   switch (getType()) {
-  case OZ_VAR_FUTURE:  
-    ((Future *)      this)->_cacRecurse(); 
+  case OZ_VAR_FAILED:  
+    ((Failed *)      this)->_cacRecurse(); 
     break;
   case OZ_VAR_OF:      
     ((OzOFVariable*) this)->_cacRecurse(); 
@@ -781,7 +1103,7 @@ void WeakStack::recurse(void)
   while (!isEmpty()) {
     pop(fut,val);
     DEREF(fut,ptr);
-    oz_bindFuture(ptr,val);
+    oz_bindReadOnly(ptr,val);
   }
 }
 
@@ -858,7 +1180,7 @@ void WeakDictionary::weakGC()
     if (t!=0 && !isGCMarkedTerm(t)) {
       numelem--;
       if (stream) {
-	if (!list) newstream=list=oz_newFuture(oz_rootBoard());
+	if (!list) newstream=list=oz_newReadOnly(oz_rootBoard());
 	// schedule key and value for later collection
 	OZ_Term p = oz_pair2(table->getKey(i),t);
 	weakReviveStack.push(p);
@@ -1035,8 +1357,10 @@ void ConstTerm::_cacConstRecurse(void) {
       if (gn) 
 	gCollectGName(gn);
 #endif
-      OZ_cacBlock(a->getGRef(),a->getGRef(),
-		  a->getPred()->getGSize());
+#ifdef S_CLONE
+            OZ_cacBlock(a->getGRef(),a->getGRef(),
+      		  a->getPred()->getGSize());
+#endif
       break;
     }
     
@@ -1204,15 +1528,16 @@ ConstTerm * ConstTerm::gCollectConstTermInline(void) {
   case Co_Builtin:
     return this;
 
+  case Co_Abstraction: {
+    Abstraction * ret = ((Abstraction *) this)->gCollect();
+    return ret;
+  }
+
     /*
      * ConstTermWithHome
      *
      */
 
-  case Co_Abstraction: 
-    sz = ((Abstraction *) this)->getAllocSize();
-    goto const_withhome;
-
   case Co_Chunk:
     sz = sizeof(SChunk);
     goto const_withhome;
@@ -1477,13 +1802,19 @@ TaskStack * TaskStack::_cac(void) {
       gCollectCode(pc);
 #endif
       (void) CodeArea::livenessX(pc,*Y);
+#if defined(DEBUG_LIVECALC) || defined(DEBUG_PRINTLIVECALC)
+      printf("Remapped XCONT's X 0x%x --> ", *Y);
+#endif
       *Y = (*Y)->_cac();
+#if defined(DEBUG_LIVECALC) || defined(DEBUG_PRINTLIVECALC)
+      printf("0x%x\n", *Y);
+#endif
     } else if (PC == C_LOCK_Ptr) {
       TaggedRef ct = makeTaggedConst((ConstTerm *) *CAP);
       oz_cacTerm(ct, ct);
       *CAP = tagged2Const(ct);
     } else if (PC == C_SET_SELF_Ptr) {
-      ConstTerm *ct = (ConstTerm *) *CAP;;
+      ConstTerm *ct = (ConstTerm *) *CAP;
       if (ct) {
 	TaggedRef ctt = makeTaggedConst(ct);
 	oz_cacTerm(ctt, ctt);
@@ -1505,10 +1836,58 @@ TaskStack * TaskStack::_cac(void) {
       oz_cacTerm(*((TaggedRef *) Y), *((TaggedRef *) Y));
       *CAP = ((RefsArray *) *CAP)->_cac();
     } else { // usual continuation
-      *Y   = (*Y)->_cac();
+#ifdef G_COLLECT
+      // Void dead G and Y registers if possible
+      int gLen = ((Abstraction *) *CAP)->cacGetPred()->getGSize();
+      int *gUsage = gUsageVector;
+      if (gLen > StaticGUsageVectorSize) {
+        // Static block not big enough 
+        gUsage = new int[gLen];
+      }
+      // Initialise G registers to not used
+      for (int i=gLen; i--;) gUsage[i]=0;
+
+      // If this frame is below a catch or lock frame then we will
+      // have already dealt with it in its real parent. We detect this
+      // by seeing if the Y registers have already been collected.
+      // Note we must still collect the Y and CAP refs so that they get 
+      // updated to the new copy.
+      if (!((*Y) && (*Y)->cacIsMarked())) {
+	
+	int yLen = (*Y)?(*Y)->getLen():0;
+        // If either Y or G registers are in scope then check their liveness
+        if (gLen || yLen) {
+#if defined(DEBUG_LIVECALC) || defined(DEBUG_PRINTLIVECALC)
+	  printf("\nG-Collect pid(%d) abstr(0x%x) -> 0x%x\n", (int) getpid(), *CAP, PC);
+	  //	  CodeArea::display(PC,50,stderr,NOCODE);
+	  //printf("\n");
+	  //ProgramCounter start=CodeArea::printDef(PC,stderr);
+	  //if (start != NOCODE) CodeArea::display(start,150,stderr,NOCODE);
+#endif
+          // Void Y registers, get liveness info for G registers
+          CodeArea::livenessGY((ProgramCounter) PC, newtop, 
+                               yLen, (*Y), 
+                               gLen, gUsage);
+
+	}
+      }
+      *CAP = ((Abstraction *) *CAP)->_cac(gLen, gUsage);
+      if (gLen > StaticGUsageVectorSize) {
+        delete [] gUsage;
+      }
+#else
+      // Clone
       TaggedRef ct = makeTaggedConst((ConstTerm *) *CAP);
       oz_cacTerm(ct, ct);
       *CAP = tagged2Const(ct);
+#endif
+#if defined(DEBUG_LIVECALC) || defined(DEBUG_PRINTLIVECALC)
+      printf("Remapped continuation's Y 0x%x --> ", *Y);
+#endif
+      *Y   = (*Y)->_cac();
+#if defined(DEBUG_LIVECALC) || defined(DEBUG_PRINTLIVECALC)
+      printf("0x%x\n", *Y);
+#endif
     }
   }
 }
@@ -2084,3 +2463,8 @@ OZ_Term * OZ_cacAllocBlock(int n, OZ_Term * frm) {
 
   return to;
 }
+
+
+
+
+
