@@ -10,8 +10,8 @@
  *    Organization or Person (Year(s))
  * 
  *  Last change:
- *    $Date: 2002/12/13 11:02:50 $ by $Author: duchier $
- *    $Revision: 1.151.2.3.2.1 $
+ *    $Date: 2004/01/22 17:50:10 $ by $Author: popow $
+ *    $Revision: 1.163 $
  * 
  *  This file is part of Mozart, an implementation 
  *  of Oz 3:
@@ -32,7 +32,6 @@
 #include "var_base.hh"
 #include "controlvar.hh"
 #include "dictionary.hh"
-#include "genhashtbl.hh"
 #include "urlc.hh"
 #include "site.hh"
 #include "mbuffer.hh"
@@ -198,21 +197,11 @@ OZ_Return raiseGeneric(char *id, char *msg, OZ_Term arg)
   return OZ_raiseDebug(makeGenericExc(id,msg,arg));
 }
 
-static void saveTerm(PickleBuffer* buf, TaggedRef t, Bool cloneCells)
-{
-  buf->marshalBegin();
-  char *version  =  MARSHALERVERSION;
-  marshalString(buf, version);
-  pickleTerm(buf, t, cloneCells);
-  buf->marshalEnd();
-  return;
-} 
-
-OZ_Return onlyFutures(OZ_Term l) {
+OZ_Return onlyReadOnlys(OZ_Term l) {
   if (oz_isNil(l)) return PROCEED;
   while (oz_isCons(l)) {
     OZ_Term f=oz_head(l);
-    if (!oz_isFuture(oz_deref(f))) {
+    if (!oz_isReadOnly(oz_deref(f))) {
       am.emptySuspendVarList();
       return PROCEED;
     }
@@ -234,8 +223,8 @@ ByteSink::putTerm(OZ_Term in, char *filename, char *header,
 
   //
   extractResources(in, cloneCells, resources, nogoods);
-  OZ_Return ret = onlyFutures(resources);
-  if (ret != PROCEED) 
+  OZ_Return ret = onlyReadOnlys(resources);
+  if (ret != PROCEED)
     return ret;
 
   //
@@ -258,7 +247,12 @@ ByteSink::putTerm(OZ_Term in, char *filename, char *header,
   PickleBuffer* pb = new PickleBuffer();
   if (textmode)
     pb->setTextmode();
-  saveTerm(pb, in, cloneCells);
+
+  //
+  pb->marshalBegin();
+  marshalString(pb, MARSHALERVERSION);
+  pickleTerm(in, pb, cloneCells);
+  pb->marshalEnd();
 
   //
   pb->saveBegin();
@@ -289,6 +283,9 @@ ByteSink::putTerm(OZ_Term in, char *filename, char *header,
     total -= len;
     OZ_Return result = putBytes(pos,len);
     if (result != PROCEED) {
+      do {
+	pb->chunkWritten();
+      } while (pb->unlinkNext(len));
       delete pb;
       return (result);
     }
@@ -389,7 +386,8 @@ OZ_Return
 saveDatum(OZ_Term in,OZ_Datum& dat)
 {
   ByteSinkDatum sink;
-  OZ_Return result = sink.putTerm(in,"UNKNOWN FILENAME","",0,NO,NO);
+  OZ_Return result = sink.putTerm(in,"UNKNOWN FILENAME","",0,
+				  NO,ozconf.pickleCells);
   if (result==PROCEED) {
     dat=sink.dat;
   } else {
@@ -435,7 +433,7 @@ OZ_BI_define(BIsave,2,0)
 {
   OZ_declareTerm(0,in);
   OZ_declareVirtualString(1,filename);
-  return saveIt(in,filename,"",0,0,NO,NO);
+  return saveIt(in,filename,"",0,0,NO,ozconf.pickleCells);
 } OZ_BI_end
 
 
@@ -444,7 +442,7 @@ OZ_BI_define(BIsaveCompressed,3,0)
   OZ_declareTerm(0,in);
   OZ_declareVirtualString(1,filename);
   OZ_declareInt(2,complevel);
-  return saveIt(in,filename,"",0,complevel,NO,NO);
+  return saveIt(in,filename,"",0,complevel,NO,ozconf.pickleCells);
 } OZ_BI_end
 
 
@@ -459,7 +457,8 @@ OZ_BI_define(BIsaveWithHeader,4,0)
   filename = strdup(filename);
   OZ_declareVS(2,header,len);
 
-  OZ_Return ret = saveIt(value,filename,header,len,compressionlevel,NO,NO);
+  OZ_Return ret = saveIt(value,filename,header,len,compressionlevel,
+			 NO,ozconf.pickleCells);
   free(filename);
   return ret;
 } OZ_BI_end
@@ -494,7 +493,7 @@ Bool pickle2text()
     return NO;
   } 
   char * s = OZ_stringToC(header,0);
-  aux = saveIt(res,"-",s,strlen(s),0,OK,NO);
+  aux = saveIt(res,"-",s,strlen(s),0,OK,ozconf.pickleCells);
   if (aux==RAISE) {
     fprintf(stderr,"Exception: %s\n",OZ_toC(am.getExceptionValue(),10,100));
     return NO;
@@ -520,15 +519,7 @@ LoadTermRet loadTerm(PickleBuffer *buf, char* &vers, OZ_Term &t)
 {
   buf->unmarshalBegin();	
 
-#ifndef USE_FAST_UNMARSHALER   
-  int error;
-  vers = unmarshalVersionStringRobust(buf, &error);
-  if (error)
-    return (CLT_NOPICKLE);
-#else
   vers = unmarshalVersionString(buf);
-#endif
-
   if (vers == 0)
     return (CLT_NOPICKLE);
 
@@ -538,13 +529,7 @@ LoadTermRet loadTerm(PickleBuffer *buf, char* &vers, OZ_Term &t)
   if (major != MARSHALERMAJOR || minor != MARSHALERMINOR)
     return (CLT_WRONGVERS);
 
-#ifndef USE_FAST_UNMARSHALER   
-  t = unpickleTermRobust(buf);
-  if (t == (OZ_Term) 0)
-    return (CLT_FORMATERR);
-#else
   t = unpickleTerm(buf);
-#endif
 
   buf->unmarshalEnd();
 
@@ -615,8 +600,6 @@ ByteSource::getTerm(OZ_Term out, const char *compname, Bool wantHeader)
 
   OZ_Return result = loadPickleBuffer(buffer, compname);
   if (result != PROCEED) {
-    buffer->dropBuffers();
-    delete buffer;
     return (result);
   }
 
@@ -693,6 +676,7 @@ ByteSource::loadPickleBuffer(PickleBuffer*& buffer, const char *filename)
     int got;
     OZ_Return result = getBytes(pos, max, got);   
     if (result != PROCEED) {
+      buffer->dropBuffers();
       delete buffer;
       DebugCode(buffer = (PickleBuffer *) 0;);
       return result;
@@ -710,6 +694,7 @@ ByteSource::loadPickleBuffer(PickleBuffer*& buffer, const char *filename)
 
   //
   if (total == 0) {
+    buffer->dropBuffers();
     delete buffer;
     DebugCode(buffer = (PickleBuffer *) 0;);
     return (raiseGeneric("bytesource:empty",
@@ -718,6 +703,7 @@ ByteSource::loadPickleBuffer(PickleBuffer*& buffer, const char *filename)
 					  oz_atom(filename)), oz_nil())));
   }
   if (checkChecksum(crc) == NO) {
+    buffer->dropBuffers();
     delete buffer;
     DebugCode(buffer = (PickleBuffer *) 0;);
     return (raiseGeneric("bytesource:crc","Checksum mismatch",
@@ -1242,3 +1228,210 @@ OZ_Term toXML(OZ_Term t,marshalFun f)
   return val;
 }
 #endif
+
+#undef PROFILE_MARSHALER
+#if defined(PROFILE_MARSHALER)
+//
+#include "byteBuffer.hh"
+#include "libdp/dpMarshaler.hh"
+
+#define TestMError(PROC, MSG) \
+  OZ_raiseErrorC("testM",3,OZ_atom(PROC),OZ_inAsList(),OZ_atom(MSG))
+
+OZ_BI_define(BIemptyBuiltin, 0, 0)
+{
+  return PROCEED;
+} OZ_BI_end
+
+// Oz value + number of iterations;
+OZ_BI_define(BItestMarshaler, 2, 0)
+{
+  OZ_declareDetTerm(0, val);
+  OZ_declareInt(1, iterations);
+  PickleBuffer* buf = new PickleBuffer();
+
+  for (int i = iterations; i--; ) {
+    OZ_Term resources, nogoods;
+
+    extractResources(val, 0, resources, nogoods);
+
+    if (!oz_isNil(resources))
+      return (TestMError("testMarshaler", "resources found!"));
+    if (!oz_isNil(nogoods))
+      return (TestMError("testMarshaler", "nogoods found!"));
+
+    buf->marshalBegin();
+    pickleTerm(val, buf, 0);
+    buf->marshalEnd();
+
+    buf->saveBegin();
+    int len; 
+    (void) buf->unlinkFirst(len);
+    do {
+      buf->chunkWritten();
+    } while (buf->unlinkNext(len));
+    buf->saveEnd();
+  }
+
+  delete buf;
+  return PROCEED;
+} OZ_BI_end
+
+// Oz value + number of iterations.
+// Note: first, the value is marshaled, which takes some time;
+OZ_BI_define(BItestUnmarshaler, 2, 0)
+{
+  OZ_declareDetTerm(0, val);
+  OZ_declareInt(1, iterations);
+
+  OZ_Term resources, nogoods;
+
+  extractResources(val, 0, resources, nogoods);
+
+  if (!oz_isNil(resources))
+    return (TestMError("testMarshaler", "resources found!"));
+  if (!oz_isNil(nogoods))
+    return (TestMError("testMarshaler", "nogoods found!"));
+
+  PickleBuffer* buf = new PickleBuffer();
+  buf->marshalBegin();
+  pickleTerm(val, buf, 0);
+  buf->marshalEnd();
+
+  for (int i = iterations; i--; ) {
+    buf->unmarshalBegin();	
+    (void) unpickleTerm(buf);
+    buf->unmarshalEnd();
+  }
+
+  // bug in 1.2.0 : just 'delete buf' doesn't delete the whole buffer;
+  buf->saveBegin();
+  int len; 
+  (void) buf->unlinkFirst(len);
+  do {
+    buf->chunkWritten();
+  } while (buf->unlinkNext(len));
+  buf->saveEnd();
+  delete buf;
+
+  return PROCEED;
+} OZ_BI_end
+
+// Oz value, number of iterations, buffer size (kb) --> total bytes written;
+OZ_BI_define(BItestDPMarshaler, 3, 1)
+{
+  OZ_declareDetTerm(0, val);
+  OZ_declareInt(1, iterations);
+  OZ_declareInt(2, bufferSize);
+
+  const int bbSize = 1024*bufferSize;
+  BYTE *bb = new BYTE[bbSize];
+  int bytes = 0;
+
+  DPMarshaler *dpm = (DPMarshaler *) new DPMarshaler;
+  ByteBuffer* buf = new ByteBuffer();
+  buf->init(bbSize, bb);
+
+  for (int i = iterations; i--; ) {
+    buf->reinit();
+    buf->marshalBegin();
+    DPMarshaler *dpmC; 
+    if ((dpmC = dpMarshalTerm(val, buf, dpm, (DSite *) 0))) {
+      do {
+	buf->marshalEnd();
+	bytes += buf->getUsed();
+	buf->reinit();
+	buf->marshalBegin();
+      } while ((dpmC = dpMarshalContTerm(buf, dpmC)));
+    }
+    buf->marshalEnd();
+    bytes += buf->getUsed();
+    dpm->reset();
+  }
+
+  delete buf;
+  delete dpm;
+  delete bb;
+  OZ_RETURN_INT(bytes);
+} OZ_BI_end
+
+// Oz value, number of iterations, buffer size (in kb);
+// What is really measured is marshaling + unmarshaling;
+OZ_BI_define(BItestDPUnmarshaler, 3, 0)
+{
+  OZ_declareDetTerm(0, val);
+  OZ_declareInt(1, iterations);
+  OZ_declareInt(2, bufferSize);
+
+  const int bbSize = 1024*bufferSize;
+  BYTE *bb = new BYTE[bbSize];
+
+  DPMarshaler *dpm = (DPMarshaler *) new DPMarshaler;
+  ByteBuffer* buf = new ByteBuffer();
+  buf->init(bbSize, bb);
+  Builder *b = new Builder;
+
+  for (int i = iterations; i--; ) {
+    dpUnmarshalerStartBatch(b);
+
+    buf->marshalBegin();
+    DPMarshaler *dpmC; 
+    buf->put(0xFF);          // Ctrl
+    buf->putInt(0xFFFFFFFF); // Ack
+    buf->putInt(0xFFFFFFFF); // Placeholder for Framesize
+    dpmC = dpMarshalTerm(val, buf, dpm, (DSite *) 0);
+    buf->marshalEnd();
+
+    buf->unmarshalBegin();
+    (void) buf->get();		// Ctrl
+    (void) buf->getInt();	// Ack
+    (void) buf->getInt();	// Framesize
+
+    while (dpmC) {
+	(void) dpUnmarshalTerm(buf, b);
+	buf->unmarshalEnd();
+
+	buf->reinit();
+
+	buf->marshalBegin();
+	buf->put(0xFF);          // Ctrl
+	buf->putInt(0xFFFFFFFF); // Ack
+	buf->putInt(0xFFFFFFFF); // Placeholder for Framesize
+	dpmC = dpMarshalContTerm(buf, dpmC);
+	buf->marshalEnd();
+
+	buf->unmarshalBegin();
+	(void) buf->get();	// Ctrl
+	(void) buf->getInt();	// Ack
+	(void) buf->getInt();	// Framesize
+    }
+
+    (void) dpUnmarshalTerm(buf, b);
+    buf->unmarshalEnd();
+    dpm->reset();
+    buf->reinit();
+  }
+
+  return (PROCEED);
+} OZ_BI_end
+
+/*
+ * also add this to e.g. modPickle.spec:
+    'emptyBuiltin'      => { in     => [],
+			      out    => [],
+			      BI     => BIemptyBuiltin},
+    'testMarshaler'     => { in     => ['+value', '+int'],
+			      out    => [],
+			      BI     => BItestMarshaler},
+    'testUnmarshaler'   => { in     => ['+value', '+int'],
+			      out    => [],
+			      BI     => BItestUnmarshaler},
+    'testDPMarshaler'   => { in     => ['+value', '+int', '+int'],
+                             out    => ['int'],
+                             BI     => BItestDPMarshaler},
+    'testDPUnmarshaler' => { in     => ['+value', '+int', '+int'],
+                             out    => [],
+                             BI     => BItestDPUnmarshaler},
+ */
+
+#endif // defined(PROFILE_MARSHALER)
