@@ -6,8 +6,8 @@
  *    Christian Schulte, 1999
  * 
  *  Last change:
- *    $Date: 2002/11/24 23:12:04 $ by $Author: duchier $
- *    $Revision: 1.44.2.1.2.1 $
+ *    $Date: 2004/04/08 15:44:11 $ by $Author: raph $
+ *    $Revision: 1.52 $
  * 
  *  This file is part of Mozart, an implementation 
  *  of Oz 3:
@@ -30,6 +30,9 @@
 #include "gentraverser.hh"
 #include "indexing.hh"
 
+#ifndef GY_CACHE_RESET_OFF 
+extern void resetLivenessCaches(void);
+#endif
 
 /*
  * Entry points for code garbage collection
@@ -990,15 +993,12 @@ void Builder::gCollect()
   }
 
   //
-  // In addition, we have to GC the BuilderIndexTable as well:
-  const int *ttsTop = getTTSTop();
-  const int *ttsBottom = getTTSBottom();
-  const int *ttsI = ttsTop;
-
-  //
-  while (--ttsI >= ttsBottom) {
-    OZ_Term &tte = getRef(*ttsI);
+  // In addition, we have to GC the BuilderRefTable<OZ_Term> as well:
+  OZ_Term *tp = valueRT->getFirstRef();
+  while (tp) {
+    OZ_Term &tte = *tp;
     oz_gCollectTerm(tte, tte);
+    tp = valueRT->getNextRef(tp);
   }
 }
 
@@ -1044,7 +1044,6 @@ void AM::doGCollect(void) {
 
 void AM::gCollect(int msgLevel)
 {
-  
   gCollectWeakDictionariesInit();
   vf.init();
   cacStack.init();
@@ -1077,11 +1076,22 @@ void AM::gCollect(int msgLevel)
   Assert(_inEqEq==FALSE);
   Assert(_rootBoard);
 
+#ifndef GY_CACHE_RESET_OFF
+  resetLivenessCaches();
+#endif
   //
   uFillNode = am.getCurrentOptVar();
   _rootBoard = _rootBoard->gCollectBoard();   // must go first!
   setCurrent(_rootBoard, _rootBoard->getOptVar());
   aritytable.gCollect();
+
+  // gCollectCodeAreaStart() decides whether or not this gc will discard
+  // CodeAreas. Since threads contain indirect pointers to CodeAreas, it
+  // must be run before threadsPool.gCollect().  Fixes a bug where a thread
+  // had the only live reference to a CodeArea.
+  // PR#953.
+  CodeArea::gCollectCodeAreaStart();
+
   threadsPool.gCollect();
 
   // mm2: Assert(isEmptySuspendVarList());
@@ -1089,9 +1099,8 @@ void AM::gCollect(int msgLevel)
 
   if (defaultExceptionHdl)
     oz_gCollectTerm(defaultExceptionHdl,defaultExceptionHdl);
-  oz_gCollectTerm(debugStreamTail,debugStreamTail);
+  oz_gCollectTerm(debugPort, debugPort);
 
-  CodeArea::gCollectCodeAreaStart();
   PrTabEntry::gCollectPrTabEntries();
   extRefs = extRefs->gCollect();
 
@@ -1133,7 +1142,7 @@ void AM::gCollect(int msgLevel)
 
   Assert(cacStack.isEmpty());
 
-  GT.gCollectGNameTable();
+  gnameTable.gCollectGNameTable();
   //   MERGECON gcPerdioFinal();
   gCollectSiteTable();
   (*gCollectPerdioFinal)();
@@ -1163,5 +1172,5 @@ void AM::gCollect(int msgLevel)
   cacStack.exit();
 
   //  malloc_stats();
-  DebugCode(uFillNode = 0;)
+  DebugCode(uFillNode = 0;);
 } // AM::gc
