@@ -7,8 +7,8 @@
  *  Copyright:
  * 
  *  Last change:
- *    $Date: 2002/10/19 12:51:47 $ by $Author: duchier $
- *    $Revision: 1.47.2.1.2.1 $
+ *    $Date: 2003/04/28 12:32:28 $ by $Author: glynn $
+ *    $Revision: 1.56 $
  * 
  *  This file is part of Mozart, an implementation 
  *  of Oz 3:
@@ -40,14 +40,19 @@
 #define CLOSE_TIMEOUT		ozconf.dpCloseTimeout
 #define WF_REMOTE_TIMEOUT	ozconf.dpWFRemoteTimeout
 #define WF_REOPEN_TIMEOUT	ozconf.dpFirewallReopenTimeout
+// one size fits all - currently cannot have different max sizes 
+// for different sites;
+#define MAX_BUFFER_SIZE		ozconf.dpMaxBufferSize
 
 #define MSG_ACK_TIMEOUT 1000
-#define MSG_ACK_LENGTH 50
+#define MSG_ACK_LENGTH 200
 
 
 #define DO_CONNECT_LOG ozconf.dpLogConnectLog
 #define DO_MESSAGE_LOG ozconf.dpLogMessageLog
 
+
+
 ComObj::ComObj(DSite *site) {
   init(site);
 }
@@ -63,6 +68,8 @@ void ComObj::init(DSite *site) {
 
   lastReceived=0;
   lastSent=0;
+  sentLrgMsg = 0;
+  receivedLrgMsg = 0;
   timer=NULL;
   reopentimer=NULL;
   closetimer=NULL;
@@ -86,7 +93,12 @@ void ComObj::init(DSite *site) {
 
 // Specifying priority -1 means accepting the default as in msgFormat.m4 and
 // should allways be used.
-void ComObj::send(MsgContainer *msgC,int priority) {
+void ComObj::send(MsgContainer *msgC)
+{
+  // Erik, hack until the prio thing has been properly reworked.
+  int priority = -1; 
+  globalSendCounter++;
+
   if(DO_MESSAGE_LOG) {
     fprintf(logfile,"send(%s %d %d %d %s)\n",
 	    mess_names[msgC->getMessageType()],
@@ -119,17 +131,23 @@ void ComObj::send(MsgContainer *msgC,int priority) {
     break;
   }
 
-  //printf("snapshot started...\n");
-  //struct timeval start, end;
-  //gettimeofday(&start, NULL);
-  msgC->takeSnapshot();
-  //gettimeofday(&end, NULL);
-  //printf("  ... snapshot time %f s\n", (end.tv_sec - start.tv_sec)+
-  //	 ((double) (end.tv_usec - start.tv_usec))/1000000);
-
-
-  Assert(priority==-1); // Have a good reason before removing this line
-  if(priority==-1) priority=msgC->getPriority();
+  // Old priority calculation. If the suplied priority is different from 
+  // -1 then use it else get the default
+  /*
+    
+    Assert(priority==-1); // Have a good reason before removing this line
+    if(priority==-1) priority=msgC->getPriority();
+  */
+  // New priority caclualtion.  Dynamicaly defined prio either sets the prio or
+  // defines that the orgin should be used. 
+  
+  if (default_mess_priority[msgC->getMessageType()] == USE_PRIO_OF_SENDER){
+    priority = msgC->getPriority();
+    //    printf("sender prio results in %d\n", priority);
+  }
+  else
+    priority =  default_mess_priority[msgC->getMessageType()] + 1;
+  
   Assert(msgC->getMessageType()>C_FIRST || msgC->getMessageType()==M_PING
 	 || (priority<5 && priority>=1));
   queues.enqueue(msgC,priority); // After check queues.hasQueued()
@@ -276,7 +294,7 @@ void ComObj::accept(TransObj *transObj) {
 
   MsgContainer *msgC=msgContainerManager->newMsgContainer(NULL);
   msgC->put_C_PRESENT(PERDIOVERSION,myDSite);
-  send(msgC,-1);
+  send(msgC);
   transObj->readyToReceive();
 }
 
@@ -311,7 +329,8 @@ void ComObj::close(CState statetobe,Bool merging) {
     comObjDone(this);
   Assert(!connectgrantrequested && connectVar==(OZ_Term) 0x45);
   queues.clear5();
-
+  // Can this be a bug? The level 5 messages are cleared, but not
+  // eventual partialy received ad sent messages.... ERIK .
   switch(statetobe) {
   case CLOSED_WF_HANDOVER:
     //    state=CLOSED_WF_HANDOVER; happens anyway!
@@ -347,7 +366,7 @@ void ComObj::preemptTransObj() {
   clearTimers(); // And set a closing timer below
   MsgContainer *msgC=msgContainerManager->newMsgContainer(NULL);
   msgC->put_C_CLOSE_HARD();
-  send(msgC,-1);
+  send(msgC);
   state=CLOSING_HARD; // State change "closes" outgoing channel, do after send
   timers->setTimer(closetimer,CLOSE_TIMEOUT,
 		   comObj_closeTimerExpired,(void *) this);
@@ -366,13 +385,13 @@ Bool ComObj::canBeFreed() {
       if(!sentclearref) {
 	MsgContainer *msgC=msgContainerManager->newMsgContainer(NULL);
 	msgC->put_C_CLEAR_REFERENCE();
-	send(msgC,-1);
+	send(msgC);
 	sentclearref=TRUE;
       }
       clearTimers();
       MsgContainer *msgC2=msgContainerManager->newMsgContainer(NULL);
       msgC2->put_C_CLOSE_WEAK();
-      send(msgC2,-1);
+      send(msgC2);
       state=CLOSING_WEAK; // State change "closes" outgoing channel do
       // after send.
       timers->setTimer(closetimer,CLOSE_TIMEOUT,
@@ -409,24 +428,24 @@ Bool ComObj::hasNeed() {
   return localRef || queues.hasNeed();
 }
 
-inline OZ_Term ComObj::createCI(int bufferSize) {
+inline OZ_Term ComObj::createCI(int bufferSize)
+{
   int time,length;
-  if(probing) // should be also when closing down!
-    time=length=0;
-  else {
-    time=MSG_ACK_TIMEOUT;
-    length=MSG_ACK_LENGTH;
-  }
+  time=MSG_ACK_TIMEOUT;
+  length=MSG_ACK_LENGTH;
   return OZ_recordInit(oz_atom("channelinfo"),
 		       oz_cons(oz_pairAI("lastReceived",lastReceived),
-			 oz_cons(oz_pairAI("msgAckTimeOut",time),
-			   oz_cons(oz_pairAI("msgAckLength",length),
-			     oz_cons(oz_pairAI("bufferSize",bufferSize),
-			        oz_cons(oz_pairAI("hasNeed",hasNeed()),
-					oz_nil()))))));
+		       oz_cons(oz_pairAI("msgAckTimeOut",time),
+		       oz_cons(oz_pairAI("msgAckLength",length),
+		       oz_cons(oz_pairAI("bufferSize",bufferSize),
+		       oz_cons(oz_pairAI("maxBufferSize",MAX_BUFFER_SIZE),
+		       oz_cons(oz_pairAI("hasNeed",hasNeed()),
+		       oz_nil())))))));
 }
 
-inline void ComObj::extractCI(OZ_Term channelinfo,int &bufferSize) {
+inline void ComObj::extractCI(OZ_Term channelinfo,
+			      int &bufferSize, int &maxBufferSize)
+{
   int index;
   int remLastReceived;
 
@@ -438,6 +457,8 @@ inline void ComObj::extractCI(OZ_Term channelinfo,int &bufferSize) {
     queues.msgAcked(remLastReceived,TRUE,FALSE);      // Resend/Ack msgs
     queues.clearRec();                                // Clear partly received
     queues.clearCont();
+    sentLrgMsg=0;
+    receivedLrgMsg=0;
     lastSent=remLastReceived; // Others requeued
 
     index=srec->getIndex(oz_atom("msgAckTimeOut"));
@@ -448,7 +469,8 @@ inline void ComObj::extractCI(OZ_Term channelinfo,int &bufferSize) {
 
     index=srec->getIndex(oz_atom("bufferSize"));
     bufferSize=oz_intToC(srec->getArg(index));
-    Assert(bufferSize==transObj->getBufferSize());
+    index=srec->getIndex(oz_atom("maxBufferSize"));
+    maxBufferSize=oz_intToC(srec->getArg(index));
 
     index=srec->getIndex(oz_atom("hasNeed"));
     remoteRef=oz_intToC(srec->getArg(index));
@@ -470,57 +492,60 @@ Bool ComObj::msgReceived(MsgContainer *msgC) {
       site!=NULL?site->getTimeStamp()->pid:-1));
 
   norm++;
+  globalRecCounter++;
+  
   
   MessageType mt=msgC->getMessageType();
   switch(mt) {
   case C_PRESENT:
-    if(state==OPENING_WF_PRESENT) {
+    if (state==OPENING_WF_PRESENT) {
       char *version;
       DSite *s1;
 
       msgC->get_C_PRESENT(version,s1);
 
-      if(strcmp(version,PERDIOVERSION)!=0 || s1!=site) {
+      if (strcmp(version,PERDIOVERSION)!=0 || s1!=site) {
 	msgContainerManager->deleteMsgContainer(msgC);
 	site->discoveryPerm();
 	site->probeFault(PROBE_PERM);
 	return FALSE;
-      }
-      else {
+      } else {
 	state=OPENING_WF_NEGOTIATE_ANS;
 	MsgContainer *newmsgC=msgContainerManager->newMsgContainer(NULL);
 	OZ_Term channelinfo;
 	channelinfo=createCI(transObj->getBufferSize());
 	newmsgC->put_C_NEGOTIATE(PERDIOVERSION,myDSite,channelinfo);
-	send(newmsgC,-1);
+	send(newmsgC);
       }
+    } else {
+      errorRec(mt);
     }
-    else errorRec(mt);
     break;
+
   case C_NEGOTIATE:
-    if(state==ANONYMOUS_WF_NEGOTIATE) {
+    if (state == ANONYMOUS_WF_NEGOTIATE) {
       char *version;
       DSite *s1;
       OZ_Term channelinfo;
+
       msgC->get_C_NEGOTIATE(version,s1,channelinfo);
       site=s1;
       transObj->setSite(site);
 
-      if(DO_CONNECT_LOG) {
+      if (DO_CONNECT_LOG) {
 	fprintf(logfile,"accept(%d %d %s)\n",
 		myDSite->getTimeStamp()->pid,
 		site!=NULL?site->getTimeStamp()->pid:0,
 		am.getEmulatorClock()->toString());
-
+	
       }
 
-      if(strcmp(version,PERDIOVERSION)!=0) {
+      if (strcmp(version,PERDIOVERSION)!=0) {
 	msgContainerManager->deleteMsgContainer(msgC);
 	site->discoveryPerm();
 	site->probeFault(PROBE_PERM);
 	return FALSE;
-      }
-      else {
+      } else {
 	// Tell the site about this comObj
 	ComObj *other=s1->setComObj(this);
 	if (other==(ComObj *) -1) {
@@ -529,10 +554,10 @@ Bool ComObj::msgReceived(MsgContainer *msgC) {
 	  // continue reading.
 	  msgContainerManager->deleteMsgContainer(msgC);
 	  comController->deleteComObj(this); 
-
 	  return FALSE;
-	}
-	else if (other!=NULL) {
+
+	  //
+	} else if (other!=NULL) {
 	  // This object is an anonymous object and the other
 	  // is to be used. What transObj to use needs to be
 	  // decided on.
@@ -543,36 +568,50 @@ Bool ComObj::msgReceived(MsgContainer *msgC) {
 	  comController->deleteComObj(this); 
 	  return ret; // !This object may not do anything more, but the trO
 	              // has to continue if it was adopted.
-	}
-	adoptCI(channelinfo);
-	state=WORKING;
-	// Whether the channel is ok now or not is to be decided by
-	// the future fault-model. For now it has to be for ports etc 
-	// to start working after temp&resource preemption.
-	if(probing) {
-	  timers->setTimer(probeFaultTimer,maxrtt,
-			   comObj_probeFault,(void *) this);
-	  timers->setTimer(probeIntervalTimer,probeinterval,
-			   comObj_sendProbePing,(void *) this);
-	  if(probeFired) {
-//  	    printf("probeOK (f %d t %d)\n",myDSite->getTimeStamp()->pid,
-//  		   site!=NULL?site->getTimeStamp()->pid:0-1);
-	    site->probeFault(PROBE_OK);
-	    probeFired=FALSE;
+
+	  //
+	} else {
+	  adoptCI(channelinfo);
+	  state=WORKING;
+	  // Whether the channel is ok now or not is to be decided by
+	  // the future fault-model. For now it has to be for ports etc 
+	  // to start working after temp&resource preemption.
+	  if (probing) {
+	    timers->setTimer(probeFaultTimer,maxrtt,
+			     comObj_probeFault,(void *) this);
+	    timers->setTimer(probeIntervalTimer,probeinterval,
+			     comObj_sendProbePing,(void *) this);
+	    if (probeFired) {
+//  	      printf("probeOK (f %d t %d)\n",myDSite->getTimeStamp()->pid,
+//  		     site!=NULL?site->getTimeStamp()->pid:0-1);
+	      site->probeFault(PROBE_OK);
+	      probeFired=FALSE;
+	    }
 	  }
 	}
 	timers->clearTimer(timer);
       }
+    } else {
+      errorRec(mt);
     }
-    else errorRec(mt);
     break;
+
   case C_NEGOTIATE_ANS:
-    if(state==OPENING_WF_NEGOTIATE_ANS) {
+    if (state==OPENING_WF_NEGOTIATE_ANS) {
       OZ_Term channelinfo;
-      int bufferSize;
+      int bufferSize, maxBufferSize;
+      int myBufferSize = transObj->getBufferSize();
       msgC->get_C_NEGOTIATE_ANS(channelinfo);
-      extractCI(channelinfo,bufferSize);
-      //      if (bufferSize<myBufferSize) // Shrink
+      extractCI(channelinfo, bufferSize, maxBufferSize);
+
+      // Handle the buffer size issue.
+      // The biggest of local and remote buffer sizes is choosen, but
+      // not exceeding either of 'maxBufferSize's;
+      bufferSize = max(bufferSize, myBufferSize);
+      bufferSize = min(bufferSize, min(MAX_BUFFER_SIZE, maxBufferSize));
+      //
+      if (bufferSize != myBufferSize)
+	transObj->setBufferSize(bufferSize);
 
       state=WORKING;
       // Whether the channel is ok now or not is to be decided by
@@ -617,7 +656,7 @@ Bool ComObj::msgReceived(MsgContainer *msgC) {
     if(state==WORKING) {
       MsgContainer *newmsgC=msgContainerManager->newMsgContainer(NULL);
       newmsgC->put_C_CLOSE_ACCEPT();
-      send(newmsgC,-1);
+      send(newmsgC);
       state=CLOSING_WF_DISCONNECT;
     }
     else if(state==CLOSING_HARD || state==CLOSING_WEAK) {
@@ -644,7 +683,7 @@ Bool ComObj::msgReceived(MsgContainer *msgC) {
 	newmsgC->put_C_CLOSE_ACCEPT();
 	newstate=CLOSED;
       }
-      send(newmsgC,-1);
+      send(newmsgC);
       clearTimers();
       state=newstate;  // State change may "close" outgoing channel.
     }
@@ -685,19 +724,36 @@ Bool ComObj::msgReceived(MsgContainer *msgC) {
 	transObj->deliver();
     }
     else errorRec(mt);
+
     break;
+
+  case C_SEND_PING_PONG:
+    {
+      int id, ctr;
+      msgC->get_C_SEND_PING_PONG(id,ctr);
+      ctr--;
+      if(ctr == 0){
+	if(id>0)ResumeSendJobb(id);
+      }
+      else{
+	MsgContainer *newmsgC=msgContainerManager->newMsgContainer(NULL);
+	newmsgC->put_C_SEND_PING_PONG(id,ctr);
+	send(newmsgC);
+      }
+      break;
+    }
   case C_CLEAR_REFERENCE: // If this is what we waited for start to close...
     if(state==WORKING||state==CLOSING_WEAK||state==CLOSING_HARD) {//in is open
       remoteRef=FALSE;
       if(state==WORKING && !hasNeed()) {
 	MsgContainer *newmsgC=msgContainerManager->newMsgContainer(NULL);
 	newmsgC->put_C_CLEAR_REFERENCE();
-	send(newmsgC,-1);
+	send(newmsgC);
 
 	// remoteRef==FALSE!
 	MsgContainer *newmsgC2=msgContainerManager->newMsgContainer(NULL);
 	newmsgC2->put_C_CLOSE_WEAK();
-	send(newmsgC2,-1);
+	send(newmsgC2);
 	state=CLOSING_WEAK; // State change "closes" outgoing channel
 	                    // after send.
 	timers->setTimer(closetimer,CLOSE_TIMEOUT,
@@ -711,7 +767,7 @@ Bool ComObj::msgReceived(MsgContainer *msgC) {
       ++lastReceived;
       remoteRef=TRUE; // Implicitly known since he is sending a message!
 
-      perdio_msgReceived(msgC);
+      perdio_msgReceived(msgC,site);
 
       if(!msgAckLength || // avoid division with zero
 	 lastReceived%msgAckLength==0)  // Time for explicit acknowledgement
@@ -729,14 +785,16 @@ Bool ComObj::msgReceived(MsgContainer *msgC) {
   return TRUE;
 }
 
-void ComObj::adoptCI(OZ_Term channelinfo){
-  int bufferSize;
+void ComObj::adoptCI(OZ_Term channelinfo)
+{
+  int bufferSize, maxBufferSize;
+  int myBufferSize = transObj->getBufferSize();
 
   retryTimeout=ozconf.dpRetryTimeFloor;
   // Whether the channel is ok now or not is to be decided by
   // the future fault-model. For now it has to be for ports etc 
   // to start working after temp&resource preemption.
-  if(probing) {
+  if (probing) {
     timers->setTimer(probeFaultTimer,maxrtt,
 		     comObj_probeFault,(void *) this);
     timers->setTimer(probeIntervalTimer,probeinterval,
@@ -749,15 +807,23 @@ void ComObj::adoptCI(OZ_Term channelinfo){
     }
   }
 
-  extractCI(channelinfo,bufferSize);
-//    if (bufferSize<myBufferSize)
-//     // shrink buffer, set myBuffersize to bufferSize, answer with bufferSize
-//    else
-//      // answer with myBufferSize
+  //
+  extractCI(channelinfo, bufferSize, maxBufferSize);
+
+  // Handle the buffer size issue.
+  // The biggest of local and remote buffer sizes is choosen, but not
+  // exceeding either of 'maxBufferSize's;
+  bufferSize = max(bufferSize, myBufferSize);
+  bufferSize = min(bufferSize, min(MAX_BUFFER_SIZE, maxBufferSize));
+  //
+  if (bufferSize != myBufferSize)
+    transObj->setBufferSize(bufferSize);
+
+  //
   MsgContainer *newmsgC=msgContainerManager->newMsgContainer(NULL);
   channelinfo=createCI(bufferSize);
   newmsgC->put_C_NEGOTIATE_ANS(channelinfo);
-  send(newmsgC,-1);
+  send(newmsgC);
 }
 
 Bool ComObj::merge(ComObj *old,ComObj *anon,OZ_Term channelinfo) {
@@ -780,11 +846,10 @@ Bool ComObj::merge(ComObj *old,ComObj *anon,OZ_Term channelinfo) {
     old->close(CLOSED,TRUE);
     goto adopt_anon;
   case OPENING_WF_NEGOTIATE_ANS:
-    if(myDSite->compareSites(site)<0) {
+    if (myDSite->compare(site) < 0) {
       old->close(CLOSED,TRUE);
       goto adopt_anon;
-    }
-    else {
+    } else {
       goto drop_anon;
     }
   case WORKING:
@@ -873,11 +938,10 @@ void ComObj::msgAcked(int num) {
 
 MsgContainer *ComObj::getNextMsgContainer(int &acknum) {
   MsgContainer *msgC=queues.getNext(state==WORKING);
-
-  if(msgC!=NULL && !msgC->checkFlag(MSG_HAS_MARSHALCONT) &&
-     msgC->getMessageType()<C_FIRST) {
-    //    queues.insertUnacked(msgC);
-    msgC->setMsgNum(++lastSent);
+  
+  if(msgC==NULL) return NULL;
+  if(!msgC->checkFlag(MSG_HAS_MARSHALCONT) && msgC->getMessageType()<C_FIRST) {
+    msgC->setMsgNum(NO_MSG_NUM);
   }
   if(DO_MESSAGE_LOG) {
     if(msgC!=NULL) 
@@ -889,7 +953,7 @@ MsgContainer *ComObj::getNextMsgContainer(int &acknum) {
 	      am.getEmulatorClock()->toString());
   }
 
-  if(probing && msgC!=NULL && msgC->getMessageType()<C_FIRST) {
+  if(probing && msgC->getMessageType()<C_FIRST) {
     msgC->setSendTime(am.getEmulatorClock());
     timers->setTimer(probeIntervalTimer,probeinterval,
 		     comObj_sendProbePing,(void *) this);
@@ -897,13 +961,16 @@ MsgContainer *ComObj::getNextMsgContainer(int &acknum) {
   if(state==WORKING && timer!=NULL) // Acknowledgement is being sent
     timers->clearTimer(timer);
   acknum=lastReceived;
+
+  // Resume any threads that are suspended due to flowcntrl. 
+  msgC->bindCntrlVar();
   return msgC;
 }
 
 void ComObj::msgSent(MsgContainer *msgC)
 {
   if (msgC->getMessageType() < C_FIRST) {
-    Assert(msgC->getMsgNum() != -1);
+    msgC->setMsgNum(++lastSent);
     queues.insertUnacked(msgC);
   } else {
     msgContainerManager->deleteMsgContainer(msgC);
@@ -911,6 +978,8 @@ void ComObj::msgSent(MsgContainer *msgC)
 }
 
 void ComObj::msgPartlySent(MsgContainer *msgC) { 
+  if (msgC->getMsgNum() == NO_MSG_NUM)
+    msgC->setMsgNum(++sentLrgMsg);
   queues.requeue(msgC);
 }
 
@@ -918,8 +987,8 @@ void ComObj::msgPartlyReceived(MsgContainer *msgC) {
   PD((TCP_INTERFACE,"---msgPartlyReceived: %s nr:%d from %d",
       mess_names[msgC->getMessageType()],lastReceived+1,
       site!=NULL?site->getTimeStamp()->pid:-1));
-  if(msgC->getMsgNum()==-1)
-    msgC->setMsgNum(lastReceived+1);
+  if(msgC->getMsgNum()==NO_MSG_NUM)
+    msgC->setMsgNum(++receivedLrgMsg);
   queues.putRec(msgC);
 }
 
@@ -1002,7 +1071,7 @@ void ComObj::installProbe(int lowerBound, int higherBound, int interval) {
     if(!probing) {
       MsgContainer *msgC=msgContainerManager->newMsgContainer(NULL);
       msgC->put_C_SET_ACK_PROP(0,0);
-      send(msgC,-1);
+      send(msgC);
     }
   }
   probing=TRUE;
@@ -1011,7 +1080,7 @@ void ComObj::installProbe(int lowerBound, int higherBound, int interval) {
 Bool ComObj::sendProbePing() {
   MsgContainer *msgC=msgContainerManager->newMsgContainer(NULL);
   msgC->put_M_PING();
-  send(msgC,-1);
+  send(msgC);
   return TRUE;
 }
 
@@ -1033,7 +1102,7 @@ Bool ComObj::sendAck(Bool timerInvoked) {
   if(!queues.hasQueued()) { // Otherwise an ack will go anyway
     MsgContainer *msgC=msgContainerManager->newMsgContainer(NULL);
     msgC->put_C_ACK();
-    send(msgC,-1);
+    send(msgC);
   }
   if(timerInvoked)
     timer=NULL;
