@@ -11,8 +11,8 @@
  *    Organization or Person (Year(s))
  * 
  *  Last change:
- *    $Date: 2002/12/06 18:35:00 $ by $Author: bruni $
- *    $Revision: 1.168.2.13.2.4 $
+ *    $Date: 2004/01/22 20:10:47 $ by $Author: duchier $
+ *    $Revision: 1.192 $
  * 
  *  This file is part of Mozart, an implementation 
  *  of Oz 3:
@@ -190,11 +190,11 @@ static long emulatorStartTime = 0;
 // observed in its test.
 // ===================================================================
 
-static int    INNER_TICKS_PER_SEC_AS_INT;
-static double INNER_TICKS_PER_SEC_AS_DOUBLE;
+static int    INNER_TICKS_PER_SEC_AS_INT = 0;
+static double INNER_TICKS_PER_SEC_AS_DOUBLE = 0.0;
 #ifdef CLK_TCK_BUG_RATIO
-static int    OUTER_TICKS_PER_SEC_AS_INT;
-static double OUTER_TICKS_PER_SEC_AS_DOUBLE;
+static int    OUTER_TICKS_PER_SEC_AS_INT = 0;
+static double OUTER_TICKS_PER_SEC_AS_DOUBLE = 0.0;
 #else
 #define OUTER_TICKS_PER_SEC_AS_INT    INNER_TICKS_PER_SEC_AS_INT
 #define OUTER_TICKS_PER_SEC_AS_DOUBLE INNER_TICKS_PER_SEC_AS_DOUBLE
@@ -216,6 +216,7 @@ unsigned int osUserTime()
   struct tms buffer;
 
   times(&buffer);
+  Assert(INNER_TICKS_PER_SEC_AS_DOUBLE>0.0); // osInit() first
   return (unsigned int)(buffer.tms_utime*1000.0/INNER_TICKS_PER_SEC_AS_DOUBLE);
 #endif
 }
@@ -235,6 +236,7 @@ unsigned int osSystemTime()
   struct tms buffer;
 
   times(&buffer);
+  Assert(INNER_TICKS_PER_SEC_AS_DOUBLE>0.0); // osInit() first
   return (unsigned int)(buffer.tms_stime*1000.0/INNER_TICKS_PER_SEC_AS_DOUBLE);
 #endif
 }
@@ -258,6 +260,7 @@ unsigned int osTotalTime()
 
   struct tms buffer;
   int t = times(&buffer) - emulatorStartTime;
+  Assert(OUTER_TICKS_PER_SEC_AS_DOUBLE>0.0); // osInit() first
   return (unsigned int) (t*1000.0/OUTER_TICKS_PER_SEC_AS_DOUBLE);
 
 #endif
@@ -316,6 +319,9 @@ void osBlockSignals(Bool check)
   sigdelset(&s,SIGINT);
   sigdelset(&s,SIGHUP);
   sigdelset(&s,SIGTERM);
+#if defined(PROFILE_EMULATOR) && defined(SIGPROF)
+  sigdelset(&s,SIGPROF);
+#endif
 
   sigprocmask(SIG_SETMASK,&s,&sOld);
 
@@ -775,16 +781,6 @@ int rawwrite(int fd, void *buf, int sz)
 
 #endif
 
-static 
-int nonBlockSelect(int nfds, fd_set *readfds, fd_set *writefds)
-{
-  struct timeval timeout;
-  timeout.tv_sec = 0;
-  timeout.tv_usec = 0;
-  return select(nfds,readfds,writefds,NULL,&timeout);
-}
-
-
 /* under windows FD_SET is not idempotent */
 #define OZ_FD_SET(i,fds) if (!FD_ISSET(i,fds)) { FD_SET(i,fds); }
 #define OZ_FD_CLR(i,fds) if (FD_ISSET(i,fds))  { FD_CLR(i,fds); }
@@ -1018,8 +1014,10 @@ void osInitSignals()
  *       Sockets                                         *
  *********************************************************/
 
-static fd_set globalFDs[2];     // mask of active read/write FDs
-static fd_set copyFDs[2];	// ... its copy for 'select()';
+static fd_set registeredFDs[2];	// mask of active read/write FDs
+static fd_set watchedFDs[2];	// ... its copy went through 'select()';
+static int numbOfWatchedFDs;	// 
+static int blockWatchedFDs;	// 
 
 int osOpenMax()
 {
@@ -1158,7 +1156,7 @@ char *osinet_ntoa(char *ip)
 }
 
 #if !defined(WINDOWS) && !defined(SUNOS_SPARC)
-int OUTER_TICKS_PER_10MS_AS_INT;
+int OUTER_TICKS_PER_10MS_AS_INT = 0;
 #endif
 
 void osInit()
@@ -1167,8 +1165,8 @@ void osInit()
 
   openMax=osOpenMax();
 
-  FD_ZERO(&globalFDs[SEL_READ]);
-  FD_ZERO(&globalFDs[SEL_WRITE]);
+  FD_ZERO(&registeredFDs[SEL_READ]);
+  FD_ZERO(&registeredFDs[SEL_WRITE]);
 
   FD_ZERO(&socketFDs);
 
@@ -1239,11 +1237,11 @@ static
 void osWatchFDInternal(int fd, int mode)
 {
   CheckMode(mode);
-  OZ_FD_SET(fd,&globalFDs[mode]); 
+  OZ_FD_SET(fd,&registeredFDs[mode]); 
   // kost@ : in the case of preventing blocking in 'select()' by
   // waiting for write permission into 'stderr' (used by task
   // checkers&handlers, see am.cc), the copy must be updated as well:
-  OZ_FD_SET(fd,&copyFDs[mode]); 
+  OZ_FD_SET(fd,&watchedFDs[mode]); 
 }
 
 
@@ -1261,14 +1259,14 @@ void osWatchAccept(int fd)
 void osClrWatchedFD(int fd, int mode)
 {
   CheckMode(mode);
-  OZ_FD_CLR(fd,&globalFDs[mode]); 
+  OZ_FD_CLR(fd,&registeredFDs[mode]); 
 }
 
 
 Bool osIsWatchedFD(int fd, int mode)
 {
   CheckMode(mode);
-  return (FD_ISSET(fd,&globalFDs[mode])); 
+  return (FD_ISSET(fd,&registeredFDs[mode])); 
 }
 
 
@@ -1278,11 +1276,93 @@ Bool osIsWatchedFD(int fd, int mode)
  */
 void osBlockSelect(unsigned int &ms)
 {
-  copyFDs[SEL_READ]  = globalFDs[SEL_READ];
-  copyFDs[SEL_WRITE] = globalFDs[SEL_WRITE];
-  (void) osSelect(&copyFDs[SEL_READ],&copyFDs[SEL_WRITE],&ms);
+  watchedFDs[SEL_READ]  = registeredFDs[SEL_READ];
+  watchedFDs[SEL_WRITE] = registeredFDs[SEL_WRITE];
+  numbOfWatchedFDs =
+    osSelect(&watchedFDs[SEL_READ], &watchedFDs[SEL_WRITE], &ms);
+  blockWatchedFDs = OK;
 }
 
+/*
+void osBusyWaitSelect(unsigned int &ms)
+{
+  int delay;
+  unsigned int currentSystemTime;
+
+  //
+  currentSystemTime = osTotalTime();
+
+  // calibrated?
+  if (ozconf.bwlIterationsPerMS == 0) {
+    unsigned int t, ref = osTotalTime();
+    int ldelay;
+
+    // skip to next tick;
+    ldelay = 0;
+    do {
+      ldelay++;
+      t = osTotalTime();
+    } while (t == ref);
+    ref = t;
+
+    // measure #iterations for 'osTotalTime()';
+    ldelay = 0;
+    do {
+      ldelay++;
+      t = osTotalTime();
+    } while (t == ref);
+    ref = t;
+
+    //
+    delay = 0;
+    do {
+      delay++;
+      watchedFDs[SEL_READ]  = registeredFDs[SEL_READ];
+      watchedFDs[SEL_WRITE] = registeredFDs[SEL_WRITE];
+      (void) osSelect(&watchedFDs[SEL_READ], &watchedFDs[SEL_WRITE],
+		      (unsigned int *) WAIT_NULL);
+      t = osTotalTime();
+    } while (t == ref);
+
+    // Now we know how many iterations are needed for both
+    // 'osTotalTime()' alone, and together with 'osSelect()'
+    // per 1ms;
+    ozconf.bwlIterationsPerMS =
+      (delay * ldelay) / ((t - ref) * (ldelay - delay));
+    // fprintf(stdout, "delay: %d\n", ozconf.bwlIterationsPerMS);
+  }
+
+  //
+  delay = ozconf.bwlIterationsPerMS;
+  if (ms == 0)
+    delay = delay * ozconf.bwlMSs;
+  else
+    delay = delay * min(ms, ozconf.bwlMSs);
+
+  //
+  do {
+    watchedFDs[SEL_READ]  = registeredFDs[SEL_READ];
+    watchedFDs[SEL_WRITE] = registeredFDs[SEL_WRITE];
+    numbOfWatchedFDs = 
+      osSelect(&watchedFDs[SEL_READ], &watchedFDs[SEL_WRITE],
+	       (unsigned int *) WAIT_NULL);
+  } while (numbOfWatchedFDs == 0 && delay--);
+
+  //
+  if (numbOfWatchedFDs) {
+    // fprintf(stdout, "bingo (%d) !\n", delay);
+    ms = osTotalTime() - currentSystemTime;
+    blockWatchedFDs = OK;
+  } else {
+    // also account for the time already spent;
+    currentSystemTime = osTotalTime() - currentSystemTime;
+    ms = ms - currentSystemTime;
+    (void) osBlockSelect(ms);
+    ms = ms + currentSystemTime;
+  }
+}
+*/
+
 /* osClearSocketErrors
  * remove the closed/failed descriptors from the fd_sets
  */
@@ -1290,10 +1370,10 @@ void osClearSocketErrors()
 {
   fd_set auxFDs[2];
 
-  /* osClrWatchedFD might change globalFDs, such that the next
-   * call to FD_ISSET on globalFDs might fail */
-  auxFDs[SEL_READ]  = globalFDs[SEL_READ];
-  auxFDs[SEL_WRITE] = globalFDs[SEL_WRITE];
+  /* osClrWatchedFD might change registeredFDs, such that the next
+   * call to FD_ISSET on registeredFDs might fail */
+  auxFDs[SEL_READ]  = registeredFDs[SEL_READ];
+  auxFDs[SEL_WRITE] = registeredFDs[SEL_WRITE];
 
   for (int i = 0; i < openMax; i++) {
     for(int mode=SEL_READ; mode <= SEL_WRITE; mode++) {
@@ -1320,7 +1400,7 @@ int osTestSelect(int fd, int mode)
     fd_set fdset, *readFDs=NULL, *writeFDs=NULL;
     FD_ZERO(&fdset);
     OZ_FD_SET(fd,&fdset);
-    
+
     if (mode==SEL_READ) {
       readFDs = &fdset;
     } else {
@@ -1330,7 +1410,8 @@ int osTestSelect(int fd, int mode)
     struct timeval timeout;
     timeout.tv_sec = 0;
     timeout.tv_usec = 0;
-    int ret = nonBlockSelect(fd+1, readFDs, writeFDs);
+    int ret = select(fd+1, readFDs, writeFDs, NULL, &timeout);
+
     if (ret >= 0 || ossockerrno() != EINTR) {
       return ret;
     }
@@ -1348,31 +1429,17 @@ static fd_set tmpFDs[2];
 /* signals are blocked */
 int osFirstSelect()
 {
- loop:
-  tmpFDs[SEL_READ]  = globalFDs[SEL_READ];
-  tmpFDs[SEL_WRITE] = globalFDs[SEL_WRITE];
-
-  int numbOfFDs = osSelect(&tmpFDs[SEL_READ], &tmpFDs[SEL_WRITE],
-			   (unsigned int *) WAIT_NULL);
-
-  if (numbOfFDs < 0) {
-    if (ossockerrno() == EINTR) goto loop;
-    if (ossockerrno() != EBADF) { /* some pipes may have been closed */
-      printfds(&tmpFDs[SEL_READ]);
-      printfds(&tmpFDs[SEL_WRITE]);
-      ozpwarning("select failed");
-    }
-    osClearSocketErrors();
-  }
-  return numbOfFDs;
+  tmpFDs[SEL_READ]  = watchedFDs[SEL_READ];
+  tmpFDs[SEL_WRITE] = watchedFDs[SEL_WRITE];
+  return (numbOfWatchedFDs);
 }
 
 Bool osNextSelect(int fd, int mode)
 {
   CheckMode(mode);
 
-  if (FD_ISSET(fd,&tmpFDs[mode])) {
-    OZ_FD_CLR(fd,&tmpFDs[mode]);
+  if (FD_ISSET(fd, &tmpFDs[mode])) {
+    OZ_FD_CLR(fd, &tmpFDs[mode]);
     return OK;
   }
   return NO;
@@ -1386,28 +1453,31 @@ Bool osNextSelect(int fd, int mode)
 // called from AM::checkIO
 int osCheckIO()
 {
- loop:
-  copyFDs[SEL_READ]  = globalFDs[SEL_READ];
-  copyFDs[SEL_WRITE] = globalFDs[SEL_WRITE];
-  
-  int numbOfFDs = osSelect(&copyFDs[SEL_READ], &copyFDs[SEL_WRITE],
-			   (unsigned int *) WAIT_NULL);
-  if (numbOfFDs < 0) {
-    if (ossockerrno() == EINTR) goto loop;
-    if (ossockerrno() != EBADF) { /* some pipes may have been closed */
-      printfds(&globalFDs[SEL_READ]);
-      printfds(&globalFDs[SEL_WRITE]);
-      ozpwarning("checkIO: select failed");
+  if (blockWatchedFDs) {
+    blockWatchedFDs = NO;
+    return (numbOfWatchedFDs);
+  } else {
+  loop:
+    watchedFDs[SEL_READ]  = registeredFDs[SEL_READ];
+    watchedFDs[SEL_WRITE] = registeredFDs[SEL_WRITE];
+
+    numbOfWatchedFDs = 
+      osSelect(&watchedFDs[SEL_READ], &watchedFDs[SEL_WRITE],
+	       (unsigned int *) WAIT_NULL);
+    if (numbOfWatchedFDs < 0) {
+      if (ossockerrno() == EINTR) goto loop;
+      if (ossockerrno() != EBADF) { /* some pipes may have been closed */
+	printfds(&registeredFDs[SEL_READ]);
+	printfds(&registeredFDs[SEL_WRITE]);
+	ozpwarning("checkIO: select failed");
+      }
+      osClearSocketErrors();
     }
-    osClearSocketErrors();
+    return (numbOfWatchedFDs);
   }
-  return numbOfFDs;
 }
 
 
-
-
-
 /*
  *  we want to kill all our children on exit *
  */
@@ -1495,8 +1565,8 @@ int ossafewrite(int fd, char *buf, unsigned int len)
 
 int osclose(int fd)
 {
-  OZ_FD_CLR((unsigned int)fd,&globalFDs[SEL_READ]);
-  OZ_FD_CLR((unsigned int)fd,&globalFDs[SEL_WRITE]);
+  OZ_FD_CLR((unsigned int)fd,&registeredFDs[SEL_READ]);
+  OZ_FD_CLR((unsigned int)fd,&registeredFDs[SEL_WRITE]);
 
 #ifdef WINDOWS
   // never close stdin on Windows, leads to problems
@@ -1649,6 +1719,7 @@ int osgetEpid()
   ticks = (ms % 1000)/10;
 #else
   struct tms buffer;
+  Assert(OUTER_TICKS_PER_10MS_AS_INT>0); // osInit() first
   do {
     ticks = times(&buffer);
   } while ((ticks-emulatorStartTime)<OUTER_TICKS_PER_10MS_AS_INT);
@@ -2177,6 +2248,7 @@ static DWORD __stdcall writerThread(void *p)
       goto loop;
   }
   CloseHandle(out);
+  // keving: why is this commented out?
   //osclose(in);
   return 0;
 }
