@@ -10,8 +10,8 @@
  *    Christian Schulte, 1997
  * 
  *  Last change:
- *    $Date: 2002/01/08 16:10:46 $ by $Author: schulte $
- *    $Revision: 1.86.2.2 $
+ *    $Date: 2004/04/07 22:33:35 $ by $Author: popow $
+ *    $Revision: 1.94 $
  * 
  *  This file is part of Mozart, an implementation 
  *  of Oz 3:
@@ -35,13 +35,14 @@
 //
 
 #include "vprops.hh"
-#include "dictionary.hh"
+#include "hashtbl.hh"
 #include "fdomn.hh"
 #include "am.hh"
 #include "os.hh"
 #include "codearea.hh"
 #include "OZCONF.h"
 #include "builtins.hh"
+#include "dictionary.hh"
 
 #include <stdarg.h>
 
@@ -60,6 +61,9 @@ enum EmulatorPropertyIndex {
   PROP_PRIORITIES_HIGH,
   PROP_PRIORITIES_MEDIUM,
   PROP_PRIORITIES,
+  // PICKLES;
+  PROP_PICKLE_CELLS,
+  PROP_PICKLE,
   // TIME
   PROP_TIME_COPY,
   PROP_TIME_GC,
@@ -147,6 +151,7 @@ enum EmulatorPropertyIndex {
   PROP_INTERNAL_PROPLOCATION,
   PROP_INTERNAL_SUSPENSION,
   PROP_INTERNAL_STOP,
+  PROP_INTERNAL_ENGINE_SUSP,
   PROP_INTERNAL_DEBUG_IP,
   PROP_INTERNAL,
   // PERDIO renamed DP
@@ -155,6 +160,7 @@ enum EmulatorPropertyIndex {
   PROP_DP_FLOWBUFFERTIME,
   PROP_DP_DEBUG,
   PROP_DP_VERSION,
+  PROP_DP_CLOCKTICK,
   PROP_DP_USEALTVARPROTOCOL,
   PROP_DP_RETRYTIMECEILING,
   PROP_DP_RETRYTIMEFLOOR,
@@ -167,6 +173,10 @@ enum EmulatorPropertyIndex {
   PROP_DP_CLOSETIMEOUT,
   PROP_DP_WFREMOTETIMEOUT,
   PROP_DP_FIREWALLREOPENTIMEOUT,
+  PROP_DP_DEFAULTBUFFERSIZE,
+  PROP_DP_DEFAULTMAXBUFFERSIZE,
+  PROP_DP_BUFFERSIZE,
+  PROP_DP_MAXBUFFERSIZE,
   PROP_DP,
   // DPTABLE
   PROP_DPTABLE_DEFAULTOWNERTABLESIZE,
@@ -180,7 +190,14 @@ enum EmulatorPropertyIndex {
   PROP_DPLOG_CONNECTLOG,
   PROP_DPLOG_MESSAGELOG,
   PROP_DPLOG,
-
+  // DPGC
+  PROP_DPGC_TIMELEASE,
+  PROP_DPGC_FRACWRC,
+  PROP_DPGC_TL_LEASETIME,
+  PROP_DPGC_TL_UPDATETIME,
+  PROP_DPGC_WRC_ALPHA,
+  PROP_DPGC,
+  
   PROP_CLOSE_TIME,
 
   PROP_OZ_STYLE_USE_FUTURES,
@@ -279,6 +296,10 @@ OZ_Term GetEmulatorProperty(EmulatorPropertyIndex prop) {
     CASE_REC(PROP_PRIORITIES,"priorities",(2,AtomHigh,AtomMedium),
 	     SET_INT(AtomHigh,ozconf.hiMidRatio);
 	     SET_INT(AtomMedium,ozconf.midLowRatio););
+    // PICKLES
+    CASE_BOOL(PROP_PICKLE_CELLS,ozconf.pickleCells);
+    CASE_REC(PROP_PICKLE,"pickle",(1,AtomCells),
+	     SET_BOOL(AtomCells,ozconf.pickleCells););
     // TIME
     CASE_INT(PROP_TIME_COPY,ozconf.timeDetailed?ozstat.timeForCopy.total:0);
     CASE_INT(PROP_TIME_GC,ozconf.timeDetailed?ozstat.timeForGC.total:0);
@@ -451,6 +472,7 @@ OZ_Term GetEmulatorProperty(EmulatorPropertyIndex prop) {
   CASE_BOOL(PROP_INTERNAL_PROPLOCATION,am.isPropagatorLocation());
   CASE_BOOL(PROP_INTERNAL_SUSPENSION,ozconf.showSuspension);
   CASE_BOOL(PROP_INTERNAL_STOP,ozconf.stopOnToplevelFailure);
+  CASE_INT(PROP_INTERNAL_ENGINE_SUSP,am.getSuspCnt());
   CASE_INT(PROP_INTERNAL_DEBUG_IP,ozconf.debugIP);
   // DP
   CASE_INT(PROP_DP_DEBUG,ozconf.debugPerdio);
@@ -468,13 +490,18 @@ OZ_Term GetEmulatorProperty(EmulatorPropertyIndex prop) {
   CASE_INT(PROP_DP_CLOSETIMEOUT,ozconf.dpCloseTimeout);
   CASE_INT(PROP_DP_WFREMOTETIMEOUT,ozconf.dpWFRemoteTimeout);
   CASE_INT(PROP_DP_FIREWALLREOPENTIMEOUT,ozconf.dpFirewallReopenTimeout);
+  CASE_INT(PROP_DP_DEFAULTBUFFERSIZE,ozconf.dpDefaultBufferSize);
+  CASE_INT(PROP_DP_DEFAULTMAXBUFFERSIZE,ozconf.dpDefaultMaxBufferSize);
+  CASE_INT(PROP_DP_BUFFERSIZE,ozconf.dpBufferSize);
+  CASE_INT(PROP_DP_MAXBUFFERSIZE,ozconf.dpMaxBufferSize);
 
   case PROP_DP_VERSION: return OZ_pair2(oz_int(PERDIOMAJOR),
 					    oz_int(PERDIOMINOR));
+  case PROP_DP_CLOCKTICK: return (oz_int(CLOCK_TICK/1000));
 
   CASE_BOOL(PROP_DP_USEALTVARPROTOCOL,ozconf.dpUseAltVarProtocol);
   CASE_REC(PROP_DP,"dp",
-	   (17,oz_atomNoDup("useAltVarProtocol"),
+	   (22,oz_atomNoDup("useAltVarProtocol"),
 	    oz_atomNoDup("seifHandler"),oz_atomNoDup("debug"),
 	    oz_atomNoDup("flowBufferSize"),oz_atomNoDup("flowBufferTime"),
 	    oz_atomNoDup("version"),oz_atomNoDup("retryTimeCeiling"),
@@ -483,7 +510,12 @@ OZ_Term GetEmulatorProperty(EmulatorPropertyIndex prop) {
 	    oz_atomNoDup("probeInterval"),oz_atomNoDup("probeTimeout"),
 	    oz_atomNoDup("openTimeout"),oz_atomNoDup("closeTimeout"),
 	    oz_atomNoDup("wfRemoteTimeout"),
-	    oz_atomNoDup("firewallReopenTimeout")),
+	    oz_atomNoDup("firewallReopenTimeout"),
+	    oz_atomNoDup("defaultBufferSize"),
+	    oz_atomNoDup("defaultMaxBufferSize"),
+	    oz_atomNoDup("bufferSize"),
+	    oz_atomNoDup("maxBufferSize"),
+	    oz_atomNoDup("clockTick")),
 	   SET_BOOL(oz_atomNoDup("useAltVarProtocol"),
 		    ozconf.dpUseAltVarProtocol);
 	   SET_BOOL(oz_atomNoDup("seifHandler"), ozconf.dpSeifHandler);
@@ -509,13 +541,22 @@ OZ_Term GetEmulatorProperty(EmulatorPropertyIndex prop) {
 		   ozconf.dpWFRemoteTimeout);
 	   SET_INT(oz_atomNoDup("firewallReopenTimeout"), 
 		   ozconf.dpFirewallReopenTimeout);
+	   SET_INT(oz_atomNoDup("defaultBufferSize"),
+		   ozconf.dpDefaultBufferSize);
+	   SET_INT(oz_atomNoDup("defaultMaxBufferSize"),
+		   ozconf.dpDefaultMaxBufferSize);
+	   SET_INT(oz_atomNoDup("bufferSize"),
+		   ozconf.dpBufferSize);
+	   SET_INT(oz_atomNoDup("maxBufferSize"),
+		   ozconf.dpMaxBufferSize);
 	   SET_REC(oz_atomNoDup("version"), OZ_pair2(oz_int(PERDIOMAJOR),
-						oz_int(PERDIOMINOR)));
+						     oz_int(PERDIOMINOR)));
+	   SET_INT(oz_atomNoDup("clockTick"), CLOCK_TICK/1000);
 	   );
   CASE_INT(PROP_DPTABLE_DEFAULTOWNERTABLESIZE,
 	   ozconf.dpTableDefaultOwnerTableSize);
   CASE_INT(PROP_DPTABLE_DEFAULTBORROWTABLESIZE,
-	   ozconf.dpTableDefaultBorrowTableSize);
+	   (1<<ozconf.dpTableDefaultBorrowTableSize));
   CASE_INT(PROP_DPTABLE_LOWLIMIT, ozconf.dpTableLowLimit);
   CASE_INT(PROP_DPTABLE_EXPANDFACTOR, ozconf.dpTableExpandFactor);
   CASE_INT(PROP_DPTABLE_BUFFER, ozconf.dpTableBuffer);
@@ -528,13 +569,14 @@ OZ_Term GetEmulatorProperty(EmulatorPropertyIndex prop) {
 	   SET_INT(oz_atomNoDup("defaultOwnerTableSize"),
 		   ozconf.dpTableDefaultOwnerTableSize);
 	   SET_INT(oz_atomNoDup("defaultBorrowTableSize"),
-		   ozconf.dpTableDefaultBorrowTableSize);
+		   (1<<ozconf.dpTableDefaultBorrowTableSize));
 	   SET_INT(oz_atomNoDup("lowLimit"), ozconf.dpTableLowLimit);
 	   SET_INT(oz_atomNoDup("expandFactor"), ozconf.dpTableExpandFactor);
 	   SET_INT(oz_atomNoDup("buffer"), ozconf.dpTableBuffer);
 	   SET_INT(oz_atomNoDup("worthwhileRealloc"), 
 		   ozconf.dpTableWorthwhileRealloc);
 	   );
+
   CASE_BOOL(PROP_DPLOG_CONNECTLOG,ozconf.dpLogConnectLog);
   CASE_BOOL(PROP_DPLOG_MESSAGELOG,ozconf.dpLogMessageLog);
   CASE_REC(PROP_DPLOG,"dpLog",
@@ -542,7 +584,23 @@ OZ_Term GetEmulatorProperty(EmulatorPropertyIndex prop) {
 	   SET_BOOL(oz_atomNoDup("connectLog"),ozconf.dpLogConnectLog);
 	   SET_BOOL(oz_atomNoDup("messageLog"),ozconf.dpLogMessageLog);
 	   );
+  
+  CASE_BOOL(PROP_DPGC_TIMELEASE,ozconf.dpUseTimeLease);
+  CASE_BOOL(PROP_DPGC_FRACWRC,ozconf.dpUseFracWRC);
+  CASE_INT(PROP_DPGC_TL_LEASETIME,ozconf.dp_tl_leaseTime);
+  CASE_INT(PROP_DPGC_WRC_ALPHA,ozconf.dp_wrc_alpha);
+  CASE_INT(PROP_DPGC_TL_UPDATETIME,ozconf.dp_tl_updateTime);
 
+  CASE_REC(PROP_DPGC,"dpGC",
+	   (5,oz_atomNoDup("useTimeLease"),oz_atomNoDup("useWRC"),oz_atomNoDup("wrc_alpha"),oz_atomNoDup("tl_leaseTime"),oz_atomNoDup("tl_updateTime")),
+	   SET_BOOL(oz_atomNoDup("useTimeLease"),ozconf.dpUseTimeLease);
+	   SET_BOOL(oz_atomNoDup("useWRC"),ozconf.dpUseFracWRC);
+	   SET_INT(oz_atomNoDup("wrc_alpha"),ozconf.dp_wrc_alpha);
+	   SET_INT(oz_atomNoDup("tl_leaseTime"),ozconf.dp_tl_leaseTime);
+	   SET_INT(oz_atomNoDup("tl_updateTime"),ozconf.dp_tl_updateTime);
+	   );
+  
+  
   CASE_INT(PROP_CLOSE_TIME,ozconf.closetime);
   CASE_BOOL(PROP_OZ_STYLE_USE_FUTURES,ozconf.useFutures);
   default:
@@ -684,8 +742,9 @@ DO_INT(F,if (INT__<1||INT__>100) {oz_typeError(1,"Int[1..100]");}; DO);
 // val is guaranteed to be determined and derefed
 OZ_Return SetEmulatorProperty(EmulatorPropertyIndex prop,OZ_Term val) {
   DEREF(val,val_ptr);
-  if (oz_isVar(val))
-    return SUSPEND;
+    if (oz_isVar(val))
+      return SUSPEND;
+    
   int      INT__;
   SRecord* REC__;
   switch (prop) {
@@ -704,6 +763,10 @@ OZ_Return SetEmulatorProperty(EmulatorPropertyIndex prop,OZ_Term val) {
     CASE_REC(PROP_PRIORITIES,
 	     SET_PERCENT(AtomHigh,ozconf.hiMidRatio);
 	     SET_PERCENT(AtomMedium,ozconf.midLowRatio););
+    // PICKLE
+    CASE_BOOL(PROP_PICKLE_CELLS,ozconf.pickleCells);
+    CASE_REC(PROP_PICKLE,
+	     SET_BOOL(AtomCells,ozconf.pickleCells););
     // GC
     CASE_NAT_DO(PROP_GC_MIN,{
       ozconf.heapMinSize=INT__/KB;
@@ -795,6 +858,15 @@ OZ_Return SetEmulatorProperty(EmulatorPropertyIndex prop,OZ_Term val) {
     CASE_NAT(PROP_DP_CLOSETIMEOUT,ozconf.dpCloseTimeout);
     CASE_NAT(PROP_DP_WFREMOTETIMEOUT,ozconf.dpWFRemoteTimeout);
     CASE_NAT(PROP_DP_FIREWALLREOPENTIMEOUT,ozconf.dpFirewallReopenTimeout);
+    CASE_NAT_DO(PROP_DP_BUFFERSIZE,{
+      if (INT__ <= ozconf.dpMaxBufferSize)
+	ozconf.dpBufferSize = INT__;
+    });
+    CASE_NAT_DO(PROP_DP_MAXBUFFERSIZE,{
+      ozconf.dpMaxBufferSize = INT__;
+      ozconf.dpBufferSize = min(ozconf.dpBufferSize, INT__);
+    });
+
     // DP    
     CASE_REC(PROP_DP,
 	     SET_NAT(AtomDebugPerdio,ozconf.debugPerdio);
@@ -821,6 +893,14 @@ OZ_Return SetEmulatorProperty(EmulatorPropertyIndex prop,OZ_Term val) {
 		      ozconf.dpWFRemoteTimeout);
 	     SET_NAT(oz_atomNoDup("firewallReopenTimeout"),
 		      ozconf.dpFirewallReopenTimeout);
+	     DO_NAT(oz_atomNoDup("bufferSize"),{
+	       if (INT__ <= ozconf.dpMaxBufferSize)
+		 ozconf.dpBufferSize = INT__;
+	     });
+	     DO_NAT(oz_atomNoDup("maxBufferSize"),{
+	       ozconf.dpMaxBufferSize = INT__;
+	       ozconf.dpBufferSize = min(ozconf.dpBufferSize, INT__);
+	     });
 	     SET_BOOL(oz_atomNoDup("useAltVarProtocol"),
 		      ozconf.dpUseAltVarProtocol););
     // DPTABLE
@@ -829,7 +909,7 @@ OZ_Return SetEmulatorProperty(EmulatorPropertyIndex prop,OZ_Term val) {
       am.setSFlag(StartGC);
       return BI_PREEMPT;});
     CASE_NAT_DO(PROP_DPTABLE_DEFAULTBORROWTABLESIZE,{
-      ozconf.dpTableDefaultBorrowTableSize=INT__;
+      ozconf.dpTableDefaultBorrowTableSize=log2ceiling(INT__);
       am.setSFlag(StartGC);
       return BI_PREEMPT;});      
     CASE_PERCENT_DO(PROP_DPTABLE_LOWLIMIT,{
@@ -851,8 +931,11 @@ OZ_Return SetEmulatorProperty(EmulatorPropertyIndex prop,OZ_Term val) {
     CASE_REC(PROP_DPTABLE,
 	     SET_NAT(oz_atomNoDup("defaultOwnerTableSize"), 
 		   ozconf.dpTableDefaultOwnerTableSize);
-	     SET_NAT(oz_atomNoDup("defaultBorrowTableSize"),
-		     ozconf.dpTableDefaultBorrowTableSize);
+	     DO_NAT(oz_atomNoDup("defaultBorrowTableSize"),{
+	       ozconf.dpTableDefaultBorrowTableSize=log2ceiling(INT__);
+	       am.setSFlag(StartGC);
+	       return BI_PREEMPT;
+	     });      
 	     SET_NAT(oz_atomNoDup("lowLimit"), ozconf.dpTableLowLimit);
 	     SET_NAT(oz_atomNoDup("expandFactor"),
 		     ozconf.dpTableExpandFactor);
@@ -868,6 +951,19 @@ OZ_Return SetEmulatorProperty(EmulatorPropertyIndex prop,OZ_Term val) {
 	     SET_BOOL(oz_atomNoDup("messageLog"),ozconf.dpLogMessageLog);
 	     );
 
+    CASE_BOOL(PROP_DPGC_TIMELEASE,ozconf.dpUseTimeLease);
+    CASE_BOOL(PROP_DPGC_FRACWRC,ozconf.dpUseFracWRC);
+    CASE_NAT(PROP_DPGC_TL_LEASETIME,ozconf.dp_tl_leaseTime);
+    CASE_NAT(PROP_DPGC_TL_UPDATETIME,ozconf.dp_tl_updateTime);
+    CASE_NAT(PROP_DPGC_WRC_ALPHA, ozconf.dp_wrc_alpha);
+    CASE_REC(PROP_DPGC,
+	     SET_BOOL(oz_atomNoDup("useTimeLease"),ozconf.dpUseTimeLease);
+	     SET_BOOL(oz_atomNoDup("useWRC"),ozconf.dpUseFracWRC);
+	     SET_NAT(oz_atomNoDup("wrc_alpha"),ozconf.dp_wrc_alpha);
+	     SET_NAT(oz_atomNoDup("tl_leaseTime"),ozconf.dp_tl_leaseTime);
+	     SET_NAT(oz_atomNoDup("tl_updateTime"),ozconf.dp_tl_updateTime);
+	     );
+    
     CASE_NAT(PROP_CLOSE_TIME,ozconf.closetime);
     CASE_BOOL_DO(PROP_STANDALONE,ozconf.runningUnderEmacs=!INT__);
     CASE_BOOL(PROP_OZ_STYLE_USE_FUTURES,ozconf.useFutures);
@@ -920,7 +1016,8 @@ OZ_Return GetProperty(TaggedRef k,TaggedRef& val)
   OzDictionary* dict;
   TaggedRef entry;
   dict = tagged2Dictionary(vprop_registry);
-  if (dict->getArg(key,entry)==PROCEED)
+  entry = dict->getArg(key);
+  if (entry) 
     if (oz_isInt(entry)) {
       entry = GetEmulatorProperty((EmulatorPropertyIndex)
 				  oz_IntToC(entry));
@@ -932,7 +1029,8 @@ OZ_Return GetProperty(TaggedRef k,TaggedRef& val)
       return PROCEED;
     }
   dict = tagged2Dictionary(system_registry);
-  if (dict->getArg(key,entry)==PROCEED) {
+  entry = dict->getArg(key);
+  if (entry) {
     val=entry; return PROCEED;
   }
   return PROP__NOT__FOUND;
@@ -952,7 +1050,8 @@ OZ_Return PutProperty(TaggedRef k,TaggedRef v)
   OzDictionary* dict;
   TaggedRef entry;
   dict = tagged2Dictionary(vprop_registry);
-  if (dict->getArg(key,entry)==PROCEED)
+  entry = dict->getArg(key);
+  if (entry)
     if (OZ_isInt(entry)) {
       return SetEmulatorProperty((EmulatorPropertyIndex)
 				 oz_IntToC(entry),v);
@@ -1015,6 +1114,9 @@ static const struct prop_entry prop_entries[] = {
   {"priorities.high",PROP_PRIORITIES_HIGH},
   {"priorities.medium",PROP_PRIORITIES_MEDIUM},
   {"priorities",PROP_PRIORITIES},
+  // pickles;
+  {"pickle.cells",PROP_PICKLE_CELLS},
+  {"pickle",PROP_PICKLE},
   // TIME
   {"time.copy",PROP_TIME_COPY},
   {"time.gc",PROP_TIME_GC},
@@ -1107,12 +1209,14 @@ static const struct prop_entry prop_entries[] = {
   {"internal.propLocation",PROP_INTERNAL_PROPLOCATION},
   {"internal.suspension",PROP_INTERNAL_SUSPENSION},
   {"internal.stop",PROP_INTERNAL_STOP},
+  {"internal.engineSusps",PROP_INTERNAL_ENGINE_SUSP},
   {"internal.ip.debug",PROP_INTERNAL_DEBUG_IP},
   // DP
   {"dp.debug",PROP_DP_DEBUG},
   {"dp.useAltVarProtocol", 
 		       PROP_DP_USEALTVARPROTOCOL},
   {"dp.version",PROP_DP_VERSION},
+  {"dp.clockTick",PROP_DP_CLOCKTICK},
   {"dp.flowBufferSize",PROP_DP_FLOWBUFFERSIZE},
   {"dp.flowBufferTime",PROP_DP_FLOWBUFFERTIME},
   {"dp.seifHandler",PROP_DP_SEIFHANDLER},
@@ -1127,6 +1231,10 @@ static const struct prop_entry prop_entries[] = {
   {"dp.closeTimeout",PROP_DP_CLOSETIMEOUT},
   {"dp.wfRemoteTimeout",PROP_DP_WFREMOTETIMEOUT},
   {"dp.firewallReopenTimeout",PROP_DP_FIREWALLREOPENTIMEOUT},
+  {"dp.defaultBufferSize",PROP_DP_DEFAULTBUFFERSIZE},
+  {"dp.defaultMaxBufferSize",PROP_DP_DEFAULTMAXBUFFERSIZE},
+  {"dp.bufferSize",PROP_DP_BUFFERSIZE},
+  {"dp.maxBufferSize",PROP_DP_MAXBUFFERSIZE},
   {"dp",PROP_DP},
   // DPTABLE
   {"dpTable.defaultOwnerTableSize",
@@ -1143,6 +1251,14 @@ static const struct prop_entry prop_entries[] = {
   {"dpLog.connectLog",PROP_DPLOG_CONNECTLOG},
   {"dpLog.messageLog",PROP_DPLOG_MESSAGELOG},
   {"dpLog",PROP_DPLOG},
+  // DPGC
+  {"dpGC.useTimeLease",PROP_DPGC_TIMELEASE },
+  {"dpGC.useWRC",PROP_DPGC_FRACWRC},
+  {"dpGC.wrc_alpha",PROP_DPGC_WRC_ALPHA},
+  {"dpGC.tl_leaseTime",PROP_DPGC_TL_LEASETIME},
+  {"dpGC.tl_updateTime",PROP_DPGC_TL_UPDATETIME},
+  {"dpGC",PROP_DPGC},
+  
   //CLOSE
   {"close.time",PROP_CLOSE_TIME},
   {0,PROP__LAST},
